"""
Static helper methods which you can use in any Lambdas.
Must be completely independent with no specific requirements.
"""

__all__ = ['validate_account_to_dashed',
           'validate_account_to_int',
           'validate_list_of_numbers_from_csv',
           'camel_case_to_underscore',
           'chunks',
           'validate_uuid4',
           'rstrip_all',
           'get_one_or_none_from_dict',
           'get_one_from_dict',
           'get_list_of_multiple_or_one_or_empty_from_dict',
           'validate_date_list_from_event_or_days_back',
           'validate_date_from_something',
           'validate_datetime_from_something',
           'validate_string_matches_datetime_format',
           'is_valid_date',
           'recursive_matches_soft',
           'recursive_matches_strict',
           'recursive_matches_extract',
           'convert_string_to_words',
           'construct_dates_from_event',
           'validate_list_of_words_from_csv_or_list',
           'first_or_none',
           'recursive_update',
           'trim_arn_to_name',
           'trim_arn_to_account',
           'make_hash',
           'to_bool'
           ]

import re
import collections
import uuid
import datetime
from datetime import timezone

from copy import deepcopy
from typing import Iterable, Callable, Dict, Mapping


def validate_account_to_dashed(account):
    """
    Validates the the provided string is in valid AdWords account format and converts it to dashed format.

    :param str account: AdWords Account
    :rtype:             str
    :return:            Dashed format
    """

    account = str(account).strip()
    if re.match("[0-9]{3}-[0-9]{3}-[0-9]{4}", account):
        return account
    elif re.match("^[0-9]{10}$", account):
        return '-'.join([str(account)[0:3], str(account)[3:6], str(account)[6:]])
    else:
        raise ValueError("Invalid account format provided: {}".format(account))


def validate_account_to_int(account):
    """
    Validates the the provided string is in valid AdWords account format and converts it to integer format.

    :param (str, int) account: AdWords Account
    :return:                   Account ID as integer
    """

    account = str(account).strip().replace('-', '')
    if re.match("^[0-9]{10}$", account):
        return int(account)
    else:
        raise ValueError("Invalid account format provided: {}".format(account))


def validate_list_of_numbers_from_csv(data):
    """
    Converts a comma separated string of numeric values to a list of sorted unique integers.
    The values that do not match are skipped.

    :param (str, iterable) data:    - str | iterable
    :return:        - list(int)
    """

    if isinstance(data, str):
        return [int(x.strip()) for x in data.split(',') if x.strip().isnumeric()]
    else:
        if isinstance(data, (int, float)):
            return [data]
        result = []
        try:
            for x in data:
                if isinstance(x, (int, float)):
                    result.append(int(x))
                elif isinstance(x, str) and x.strip().isnumeric():
                    result.append(int(x.strip()))
        except TypeError:
            pass
        return result


def validate_uuid4(uuid_string):
    """
    Validate that a UUID string is in
    fact a valid uuid4.
    Happily, the uuid module does the actual
    checking for us.
    It is vital that the 'version' kwarg be passed
    to the UUID() call, otherwise any 32-character
    hex string is considered valid.
    """

    try:
        val = uuid.UUID(uuid_string, version=4)
    except ValueError:
        # If it's a value error, then the string
        # is not a valid hex code for a UUID.
        return False

    # If the uuid_string is a valid hex code,
    # but an invalid uuid4,
    # the UUID.__init__ will convert it to a
    # valid uuid4. This is bad for validation purposes.


def camel_case_to_underscore(name):
    """
    Converts attribute to string and formats it as underscored.

    :param name:    - str   -   CamelCase string (or something convertable to CamelCase with __str__() method.
    :return:        - str   -   underscore_formatted_value
    """

    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', str(name))
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def chunks(l, n):
    """Yield successive n-sized chunks from l."""

    for i in range(0, len(l), n):
        yield l[i:i + n]


def rstrip_all(input, patterns):
    """
    Strips all of the patterns from the right of the input. Order and spaces do not matter.

    :param input:       - str                   - String to modify
    :param patterns:    - list|set|tuple|str    - Pattern[-s] to remove.
    :return:            - str
    """

    if isinstance(patterns, str):
        regex = re.compile("({}$)".format(re.escape(patterns)))
    else:
        regex = re.compile("({}$)".format("$|".join(map(re.escape, patterns))))

    if isinstance(patterns, str):
        patterns = [patterns]
    if not isinstance(patterns, (list, set, tuple)) or not all(isinstance(x, str) for x in patterns):
        raise ValueError("Patterns for rstrip_all() are supposed to be string or iterable of strings")
    rabbit = input.strip()
    r = regex.sub('', rabbit)
    if not r == rabbit:
        # Go recursive in case we stripped smth in this iteration.
        return rstrip_all(r, patterns)
    else:
        # If nothing left to change, return the rabbit.
        return rabbit


def get_one_or_none_from_dict(input, name, vtype=None):
    """
    Extracts object by 'name' from the 'input'.
    Tries also plural name in case not found by single 'name'.
    In case found an iterable by plural name, validates that it has one or zero values in it.
    If vtype is specified, tries to convert result to it.

    :param dict input:  Input dictionary. Event of Lambda for example.
    :param str name:    Name of attribute (in singular form).
    :param type vtype:  Type to be converted to. Must be callable. Tested types: str, int, float
    :return:            - instance of vtype | something else | None

    :raises ValueError: In all cases something is wrong.
    """

    if not isinstance(input, dict):
        raise ValueError("'input' attribute must be a dict. Received: {}".format(type(input)))

    if not isinstance(name, str):
        raise ValueError("'name' attribute must be a str. Received: {}".format(type(name)))


    def convert(obj, t):
        return obj if not t else t(obj)


    # Best case scenario. :)
    result = input.get(name)
    if result:
        return convert(result, vtype)

    # if not result, try to search for plural
    results = input.get(name + 's')
    if not results:
        return None

    # If we found some results with plural name, we make sure it is iterable and has one or zero value.
    if isinstance(results, (list, tuple, set)):
        if len(results) > 1:
            raise ValueError("More than one {}s found in input.".format(name))
        else:
            return convert(results[0], vtype)
    elif results:
        raise ValueError("Some not-iterable '{}s' found in input: {}".format(name, str(type(result))))


def get_one_from_dict(input, name, vtype=None):
    """
    Extracts object by 'name' from the 'input'.
    Tries also plural name in case not found by single 'name'.
    In case found an iterable by plural name, validates that it has exactly one value in it.
    If vtype is specified, tries to convert result to it.

    :param input:       - dict       - Input dictionary. Event of Lambda for example.
    :param name:        - str        - Name of attribute (in singular form).
    :param vtype:       - type       - Type to be converted to. Must be callable. Tested types: str, int, float
    :return:            - instance of vtype | something else | None
    :raises ValueError:  - In all cases something is wrong.
    """

    result = get_one_or_none_from_dict(input, name, vtype)
    if result:
        return result
    else:
        raise ValueError("Did not find any value {} in the input {}".format(name, input))


def get_list_of_multiple_or_one_or_empty_from_dict(input, name, vtype=None):
    """
    Extracts objects by 'name' from the 'input' and returns as a list.
    Tries both plural and singular names from the input.
    If vtype is specified, tries to convert each of the elements in the result to this type.

    :param input:       - dict       - Input dictionary. Event of Lambda for example.
    :param name:        - str        - Name of attribute (in plural form).
    :param vtype:       - type       - Type to be converted to. Must be callable. Tested types: str, int, float
    :return:            - list       - List of vtypes, or list of whatever was in input, or empty list.
    :raises ValueError:  In all cases something is wrong.
    """

    if not isinstance(input, dict):
        raise ValueError("'input' attribute must be a dict. Received: {}".format(type(input)))

    if not isinstance(name, str):
        raise ValueError("'name' attribute must be a str. Received: {}".format(type(name)))


    def convert(obj, t):
        return obj if not t else t(obj)


    results = input.get(name) or input.get(name.rstrip('s'))
    if not results:
        return []

    # Wrap to list if not there yet.
    if not isinstance(results, (list, tuple, set)):
        results = [results]
    else:
        results = list(results)

    # Apply vtype convertion if required.
    return [convert(x, vtype) for x in results]


def validate_date_list_from_event_or_days_back(input, days_back=0, key_name='date_list'):
    """
    Takes from input the date_list and extracts date_list. Validates and converts to datetime.date.
    Input should have date_list as list of strings or comma-separated string.

    * Format:     ``YYYY-MM-DD``
    * Examples:

    .. code-block:: python

       ['2018-01-01', '2018-02-01']
       '2018-01-01, 2018-02-01'

    :param dict input:      This is supposed to be your whole Lambda event.
    :param int days_back:   Optional Number of days to take back from today.
                            Ex: days_back=1 is yesterday. Default: today.
    :param str key_name:    Optional custom name of key to extract from 'input'.
    :return:                list(datetime.date)
    """

    date_list = input.get(key_name, '')

    if not date_list:
        return [datetime.date.today() - datetime.timedelta(days=days_back)]

    if not isinstance(date_list, (list, set, tuple)):
        date_list = str(date_list).split(',')
    return [datetime.datetime.strptime(x.strip(), '%Y-%m-%d').date() for x in date_list]


def validate_datetime_from_something(d):
    """
    Converts the input `d` to datetime.datetime.

    :param d: Some input. Supported types:
                * datetime.datetime
                * datetime.date
                * int - Epoch or Epoch milliseconds
                * float - Epoch or Epoch milliseconds
                * str (YYYY-MM-DD)
                * str (YYYY-MM-DD HH:MM:SS)
    :return: Transformed `d`
    :rtype: datetime.datetime
    :raises: ValueError
    """

    mutators = [
        (datetime.datetime, lambda x: x),
        (datetime.date, lambda x: datetime.datetime.combine(x, datetime.datetime.min.time())),
        ((int, float), lambda x: datetime.datetime.fromtimestamp(x)
        if x < datetime.datetime(datetime.MAXYEAR, 12, 31).timestamp()
        else datetime.datetime.fromtimestamp(x / 1000, tz=timezone.utc)),
        (str, lambda x: datetime.datetime.strptime(d, '%Y-%m-%d')
        if len(d) == 10 else datetime.datetime.strptime(d[:19], '%Y-%m-%d %H:%M:%S'))
    ]

    for mutator in mutators:
        if isinstance(d, mutator[0]):
            return mutator[1](d)

    raise ValueError("Some unconvertable type for datetime validation: {}".format(d))


def validate_date_from_something(d):
    """
    Convert valid input to datetime.date() or raise either AttributeError or ValueError.

    :param d: Some input. Supported types:
                * datetime.datetime
                * datetime.date
                * int - Epoch or Epoch milliseconds
                * float - Epoch or Epoch milliseconds
                * str (YYYY-MM-DD)
                * str (YYYY-MM-DD HH:MM:SS)
    :return: Transformed `d`
    :rtype: datetime.date
    :raises: ValueError
    """

    return validate_datetime_from_something(d).date()


def validate_string_matches_datetime_format(date_str, date_format, field_name='date'):
    """
    Validate string, make sure it's of the given datetime format

    :param str date_str: a date or time or both, Example: '2018/09/16'
    :param str date_format: datetime format, that is acceptable for datetime.strptime. Example: '%Y/%m/%d'
            (https://docs.python.org/3.6/library/datetime.html#strftime-and-strptime-behavior)
    :param str field_name: name of the field (for the error)
    :raises: ValueError
    """

    if not (date_str and isinstance(date_str, str)):
        raise ValueError(f"Bad input for {field_name} ({date_str}), must be a string")
    if not (date_format and isinstance(date_format, str)):
        raise ValueError(f"Bad input for format ({date_format}), must be a string")

    try:
        datetime.datetime.strptime(date_str, date_format)
    except ValueError:
        raise ValueError(f"Incorrect format of {field_name} ({date_str}), should be {date_format}")


def is_valid_date(date_str, date_formats):
    """
    Validate string to be at least one of the given datetime formats.

    :param str date_str: a date or time or both, Example: '2018/09/16'
    :param list date_formats: List of datetime format, that is acceptable for datetime.strptime. Example: '%Y/%m/%d'
    :rtype: bool
    :return: True if the date string is valid for any of the datetime formats, False otherwise.
    """

    for date_format in date_formats:
        try:
            validate_string_matches_datetime_format(date_str, date_format)
            return True

        except ValueError as err:
            continue

    return False


def recursive_matches_soft(src, key, val, **kwargs):
    """
    Searches the 'src' recursively for nested elements provided in 'key' with dot notation.
    In case some levels are iterable (list, tuple) it checks every element.
    In case the full path is inaccessible returns False.
    If any of the elements addressed by 'key' matches the 'val' - Bingo! Return True.

    You might also be interested in recursive_exists_strict() helper.

    :param dict src:    Input dictionary. Can contain nested dictionaries and lists.
    :param str key:     Path to search with dot notation.
    :param any val:     Value to match in some elements specified by path.

    In order to check not just that some element exists, but to check for duplicates, you might want to use
    optional 'exclude' attributes. If attributes are specified and the last level element following the path
    (dot notation) will have a key-value, the check for the main key-value will be skipped.
    See unittests to understand the bahaviour better.

    :param str exclude_key:     Key to check in last level element to exclude.
    :param srt exclude_val:     Value to match in last level element to exclude.

    :rtype: bool
    """

    if any([x in kwargs for x in ['exclude_key', 'exclude_val']]) \
            and not all([x in kwargs for x in ['exclude_key', 'exclude_val']]):
        raise AttributeError("If you use 'exclude' attributes you must specify both 'exclude_key' and 'exclude_val'")

    path_elements = key.split('.')
    # logging.debug("Invoked func: ", src, key, path_elements)

    # if src is iterable: iterate recursively
    if isinstance(src, (list, tuple)):
        return any(recursive_matches_soft(element, key, val, **kwargs) for element in src)

    # We should try to dig deeper.
    elif len(path_elements) > 1:
        try:
            if recursive_matches_soft(src[path_elements[0]], '.'.join(path_elements[1:]), val, **kwargs):
                return True
        except KeyError:
            pass

    # Last level of digging
    elif len(path_elements) == 1:
        try:
            if kwargs.get('exclude_key') and src[kwargs['exclude_key']] == kwargs['exclude_val']:
                # logging.debug("Skipping element because it matches exclude parameters.")
                return False
        except KeyError:
            pass  # There is a chance that the exclude key is simply missing. We ignore it then.
        try:
            return src[key] == val
        except (KeyError, TypeError):
            pass
    else:
        raise RuntimeError("Your function is stupid")

    # If nothing found we return False
    return False


def recursive_matches_strict(src, key, val, **kwargs):
    """
    Searches the 'input' recursively for nested elements provided in 'key' with dot notation.
    In case some levels are iterable (list, tuple) it checks every element.
    In case the full path is inaccessible raises AttributeError or KeyError.

    :param dict src:    Input dictionary. Can contain nested dictionaries and lists.
    :param str key:     Path to search with dot notation.
    :param any val:     Value to match in some elements specified by path.

    :rtype: bool
    """

    if any([x in kwargs for x in ['exclude_key', 'exclude_val']]) \
            and not all([x in kwargs for x in ['exclude_key', 'exclude_val']]):
        raise AttributeError("If you use 'exclude' attributes you must specify both 'exclude_key' and 'exclude_val'")

    path_elements = key.split('.')

    # if src is iterable: iterate
    if isinstance(src, (list, tuple)):
        return any(recursive_matches_strict(x, key, val, **kwargs) for x in src)
    elif len(path_elements) > 1:
        return recursive_matches_strict(src[path_elements[0]], '.'.join(path_elements[1:]), val, **kwargs)
    elif len(path_elements) == 1:
        try:
            if kwargs.get('exclude_key') and src[kwargs['exclude_key']] == kwargs['exclude_val']:
                # logging.debug("Skipping element because it matches exclude parameters.")
                return False
        except KeyError:
            pass  # There is a chance that the exclude key is simply missing. We ignore it then.
        return src[key] == val
    else:
        raise RuntimeError("Your function is stupid", src, key, val)


def recursive_matches_extract(src, key, **kwargs):
    """
    Searches the 'src' recursively for nested elements provided in 'key' with dot notation.
    In case some levels are iterable (list, tuple) it checks every element in it till finds it.

    Returns the first found element or None.
    In case the full path is inaccessible also returns None.

    If you are just checking if some elements exist, you might be interested in
    recursive_exists_strict() or recursive_exists_soft() helpers.

    .. :warninig:
        Please be aware that this method doesn't not check for duplicates in iterable elements on neither
        level during extraction.

    :param dict src:    Input dictionary. Can contain nested dictionaries and lists.
    :param str key:     Path to search with dot notation.

    In order to filter out some specific elements, you might want to use the optional 'exclude' attributes.
    If attributes are specified and the last level element following the path
    (dot notation) will have a key-value, the check for the main key-value will be skipped.
    See unittests to understand the bahaviour better.

    :param str exclude_key:     Key to check in last level element to exclude.
    :param str exclude_val:     Value to match in last level element to exclude.

    :return:    Value from structure extracted by specified path
    """

    if any([x in kwargs for x in ['exclude_key', 'exclude_val']]) \
            and not all([x in kwargs for x in ['exclude_key', 'exclude_val']]):
        raise AttributeError("If you use 'exclude' attributes you must specify both 'exclude_key' and 'exclude_val'")

    path_elements = key.split('.')
    # logging.debug("Invoked func: ", src, key, path_elements)

    # if src is iterable: iterate recursively
    if isinstance(src, (list, tuple)):
        for element in src:
            v = recursive_matches_extract(element, key, **kwargs)
            if v:
                return v

    # We should try to dig deeper.
    elif len(path_elements) > 1:
        try:
            return recursive_matches_extract(src[path_elements[0]], '.'.join(path_elements[1:]), **kwargs)
        except KeyError:
            pass

    # Last level of digging
    elif len(path_elements) == 1:
        try:
            if kwargs.get('exclude_key') and src[kwargs['exclude_key']] == kwargs['exclude_val']:
                # logging.debug("Skipping element because it matches exclude parameters.")
                return None
        except KeyError:
            pass  # There is a chance that the exclude key is simply missing. We ignore it then.
        return src.get(key)
    else:
        raise RuntimeError("Your function is stupid")

    # If nothing found we return False
    return None


def convert_string_to_words(string):
    """
    Convert string to comma separated words.

    :param  str string:     String to convert into words.
    :rtype: str
    :return: Comma separated words.
    """

    if not isinstance(string, str):
        raise TypeError(f"Input must be string, got {type(string)}")

    return re.sub('\s+', ',', string.lower().strip())


def construct_dates_from_event(event: dict) -> tuple:
    """
    Processes given event dictionary for start and end points of time. Otherwise takes the default settings.

    The end date of the period may be specified as `en_date` in the event. The default value is today.

    Also the `event` should have either `st_date` or `days_back` numeric parameter.
    If provided the days_back it will be substracted from end date.

    Both `st_date` and `en_date` might be either `date`, `datetime` or `string` (`'YYYY-MM-DD'`) types.
    In case of `datetime`, the hours/minutes/etc are ignored.

    :param dict event:  Lambda payload.
    :return:            start_date, end_date    as datetime.date
    """

    en_date = validate_date_from_something(event.get('en_date', datetime.date.today()))
    st_date = event.get('st_date')
    days_back = event.get('days_back')

    if st_date and days_back:
        raise AttributeError(f"construct_dates_from_event() doesn't allow st_date and days_back simultaneously")

    if not st_date and not days_back:
        raise AttributeError(f"construct_dates_from_event() expects either st_date or days_back")

    if days_back:
        st_date = en_date - datetime.timedelta(days=int(days_back))
    else:
        st_date = validate_date_from_something(st_date)
        assert st_date < en_date, "Start date must be earlier than end date."

    return st_date, en_date


def validate_list_of_words_from_csv_or_list(data: (str, list)) -> list:
    """
    Splits a CSV string to list of stripped words.
    In case the `data` is already a list of strings - splits it's elements and flattens the result.

    All resulting elements must be single words, if any of the elements contains spaces (i.e. multiple words)
    the validation fails with `ValueError`.

    :param data:    CSV string of list of strings (possibly CSV themselves)
    :return:        List of stripped and split words
    """


    def split_csv(row):
        if not isinstance(row, str):
            raise TypeError(f"Unsupported type of data for validate_list_of_words_from_csv_or_list(): {data}")

        return [x.strip() for x in row.split(',')]


    result = []
    if isinstance(data, (list, tuple, set)):
        for element in data:
            result.extend(split_csv(element))
    else:
        result = split_csv(data)

    if any(' ' in x for x in result):
        raise ValueError(f"data for validate_list_of_words_from_csv_or_list() should be csv of WORDS or list: {data}")

    return result


def first_or_none(items: Iterable, condition: Callable = None):
    """
    Return first element in iterable to match condition or None
    """

    if not condition:
        def condition(*args, **kwargs):
            return True

    for item in items:
        if condition(item):
            return item

    return None


def recursive_update(d: Dict, u: Mapping) -> Dict:
    """
    Recursively updates the dictionary `d` with another one `u`.
    Values of `u` overwrite in case of type conflict.

    List, set and tuple values of `d` and `u` are merged, preserving only unique values. Returned as List.
    """

    new = deepcopy(d)

    for k, v in u.items():
        if isinstance(v, collections.Mapping) and isinstance(d.get(k), (collections.Mapping, type(None))):
            new[k] = recursive_update(d.get(k, {}), v)

        elif isinstance(v, (set, list, tuple)):
            if isinstance(d.get(k), (set, list, tuple)):
                # Merge lists of uniques. I really want this helper to eat anything and return what it should. :)
                nv = list(set(d[k])) + list(v)
                try:
                    # The types of values in list could be unhashable, so it is not that easy filter uniques.
                    new[k] = list(set(nv))
                except TypeError:
                    # In this case we just merge lists as is.
                    new[k] = nv
            else:
                new[k] = v
        else:
            new[k] = v

    return new


def trim_arn_to_name(arn: str) -> str:
    """
    Extract just the name of function from full ARN. Supports versions, aliases or raw name (without ARN).

    More information about ARN Format:
    https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#genref-arns
    """

    # Special handling for super global services (e.g. S3 buckets)
    if arn.count(':') < 6 and '/' not in arn:
        return arn.split(':')[-1]

    # Seems a little messy, but passes more/less any test of different ARNs we tried.
    pattern = "(arn:aws:[0-9a-zA-Z-]{2,20}:[0-9a-zA-Z-]{0,12}:[0-9]{12}:[0-9a-zA-Z-]{2,20}[:/])?" \
              "(?P<name>[0-9a-zA-Z_=,.@-]*)(:)?([0-9a-zA-Z$]*)?"

    return re.search(pattern, arn).group('name')


def trim_arn_to_account(arn: str) -> str:
    """
    Extract just the ACCOUNT_ID from full ARN. Supports versions, aliases or raw name (without ARN).

    More information about ARN Format:
    https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html#genref-arns
    """

    # Seems a little messy, but passes more/less any test of different ARNs we tried.
    pattern = "(arn:aws:[0-9a-zA-Z-]{2,20}:[0-9a-zA-Z-]{0,12}:)?(?P<acc>[0-9]{12})(:[0-9a-zA-Z-]{2,20}[:/])?" \
              "(?P<name>[0-9a-zA-Z_=,.@-]*)(:)?([0-9a-zA-Z$]*)?"

    return re.search(pattern, arn).group('acc')


def make_hash(o):
    """
    Makes a hash from a dictionary, list, tuple or set to any level, that contains
    only other hashable types (including any lists, tuples, sets, and
    dictionaries).

    Original idea from this user:
    https://stackoverflow.com/users/660554/jomido

    Plus some upgrades to work with sets and dicts having different types of keys appropriately.
    See source unittests of this function for some more details.
    """

    if isinstance(o, (tuple, list)):
        return tuple([make_hash(e) for e in o])

    # Set should be sorted (by hashes of elements) before returns
    elif isinstance(o, set):
        return tuple(sorted([make_hash(e) for e in o]))

    elif not isinstance(o, dict):
        return hash(o)

    # We are left with a dictionary
    new_o = dict()
    for k, v in o.items():
        # hash both keys and values to make sure types and order doesn't affect.
        new_o[make_hash(k)] = make_hash(v)

    return hash(tuple(frozenset(sorted(new_o.items()))))


def to_bool(val):
    if isinstance(val, (bool, int, float)):
        return bool(val)
    elif isinstance(val, str):
        if val.lower() in ['true', '1']:
            return True
        elif val.lower() in ['false', '0']:
            return False
    raise Exception(f"Can't convert unexpected value to bool: {val}, type: {type(val)}")



import json
import logging

from sosw.app import Processor
from typing import Dict

__author__ = "Nikolay Grishchenko"
__email__ = "dev@bimpression.com"
__version__ = "0.1"
__license__ = "MIT"
__status__ = "Production"


__all__ = ['Worker']

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class Worker(Processor):
    """
    We recommend that you inherit your core Processor from this class in Lambdas that are orchestrated by `sosw`.

    The ``__call__`` method is supposed to accept the ``event`` of the Lambda invocation.
    This is a dictionary with the payload received in the lambda_handler during invocation.

    Worker has all the common methods of :ref:`Processor` and tries to mark task as completed if received
    ``task_id`` in the ``event``.
    """

    DEFAULT_CONFIG = {
        'init_clients': ['lambda'],
        'sosw_worker_assistant_lambda': 'sosw_worker_assistant'
    }

    # these clients will be initialized by Processor constructor
    lambda_client = None

    def __call__(self, event: Dict):
        """
        You can either call super() at the end of your child function or completely overwrite this function.
        """

        # Mark the task as completed in DynamoDB if the event had task_id.
        try:
            self.mark_task_as_completed(event.get('task_id'))
        except Exception:
            logger.exception(f"Failed to call WorkerAssistant for event {event}")
            pass

        super().__call__(event)


    def mark_task_as_completed(self, task_id: str):
        """ Call worker assistant lambda and tell it to close task """

        if not self.lambda_client:
            self.register_clients(['lambda'])

        worker_assistant_lambda_name = self.config.get('sosw_worker_assistant_lambda', 'sosw_worker_assistant')
        payload = {
            'action': 'mark_task_as_completed',
            'task_id': task_id
        }
        payload = json.dumps(payload)

        lambda_response = self.lambda_client.invoke(
                FunctionName=worker_assistant_lambda_name,
                InvocationType='Event',
                Payload=payload
        )
        logger.debug(f"mark_task_as_completed response: {lambda_response}")



__all__ = ['Scheduler']

__author__ = "Nikolay Grishchenko"
__email__ = "dev@bimpression.com"
__version__ = "0.1"
__license__ = "MIT"
__status__ = "Development"

import datetime
import json
import logging
import os
import re
import time

from collections import Iterable
from copy import deepcopy
from typing import List, Set, Tuple, Union, Optional, Dict

from sosw.app import Processor, LambdaGlobals
from sosw.components.helpers import get_list_of_multiple_or_one_or_empty_from_dict, trim_arn_to_name, chunks
from sosw.components.siblings import SiblingsManager
from sosw.managers.task import TaskManager


logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


def single_or_plural(attr):
    """ Simple function. Gives versions with 's' at the end and without it. """
    return list(set([attr, attr.rstrip('s'), f"{attr}s"]))


def plural(attr):
    """ Simple function. Gives plural form with 's' at the end. """
    return f"{attr.rstrip('s')}s"


class InvalidJob(ValueError):
    pass


class Scheduler(Processor):
    """
    Scheduler is converting business jobs to one or multiple Worker tasks.
    """

    DEFAULT_CONFIG = {
        'init_clients':    ['Task', 's3', 'Sns', 'Siblings'],
        'task_config':     {
            'labourers': {
                # 'some_function': {
                #     'arn': 'arn:aws:lambda:us-west-2:000000000000:function:some_function',
                #     'max_simultaneous_invocations': 10,
                # }
            },
        },
        's3_prefix':       'sosw/scheduler',
        'queue_file':      'tasks_queue.txt',
        'queue_bucket':    'autotest-bucket',
        'shutdown_period': 60,
        'rows_to_process': 50,
        'job_schema':      {
            'chunkable_attrs': [
                # ('section', {}),
                # ('store', {}),
                # ('product', {}),
            ]
        }
    }

    # these clients will be initialized by Processor constructor
    task_client: TaskManager = None
    siblings_client: SiblingsManager = None
    s3_client = None
    sns_client = None
    base_query = ...


    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.set_queue_file()

        self.chunkable_attrs = list([x[0] for x in self.config['job_schema']['chunkable_attrs']])
        assert not any(x.endswith('s') for x in self.chunkable_attrs), \
            f"We do not currently support attributes that end with 's'. " \
                f"In the config you should use singular form of attribute. Received from config: {self.chunkable_attrs}"


    def __call__(self, event):
        """
        Process an event.

        :param dict event: event data
        """

        job = self.extract_job_from_payload(event)

        # If called as sibling
        if 'file_name' in job:
            self.set_queue_file(job['file_name'])

        # else construct new data file
        else:
            self.parse_job_to_file(job)

        self.process_file()

        super().__call__(event)


    def parse_job_to_file(self, job: Dict):
        """
        Splits the Job to multiple tasks and writes them down in self.local_queue_file.

        :param dict job:    Payload from Scheduled Rule.
                            Should be already parsed from whatever payload to dict and contain the raw `job`
        """

        if os.path.isfile(self.local_queue_file):
            logger.critical(f"The current Lambda container is already having some unprocessed file. "
                            f"You probably did not clean it correctly after processing. "
                            f"Should probably just clean the file now, but during beta version we raise an stop "
                            f"processing new ones.")
            raise RuntimeError(f"The current Lambda container is already having some unprocessed file.")

        labourer = self.task_client.get_labourer(labourer_id=job.pop('lambda_name'))

        # In case there is not chunking required, we just schedule `task` directly from the `job`.
        if not all([self.chunkable_attrs, self.needs_chunking(plural(self.chunkable_attrs[0]), job)]):
            data = [{'labourer_id': labourer.id, **job}]

        # Else there is much more logic how to chunk the job to tasks.
        else:
            data = self.construct_job_data(job, skeleton={'labourer_id': labourer.id})

        with open(self.local_queue_file, 'w') as f:
            for row in data:
                f.write(f"{json.dumps(row)}\n")


    # def create_tasks(self, labourer: Labourer, data: List):
    #     """
    #     Iterate tasks from `data` and queue them as new tasks for `labourer`.
    #     """
    #
    #     for task in data:
    #         self.task_client.create_task(labourer=labourer, payload=task)

    def validate_list_of_vals(self, data: Union[list, set, tuple, Dict]) -> list:
        """
        Supported resulting values: str, int, float.

        Expects a simple iterable of supported values or a dictionary with values = None.
        The keys then are treated as resulting values and if they validate, are returned as a list.
        """

        if isinstance(data, (list, set, tuple)):
            if not all(isinstance(v, (str, int, float)) for v in data):
                raise InvalidJob(f"Job has values with embedded data, but chunking is not requested. "
                                 f"You should either append 'isolate_ATTRIBUTE' flag or make values appropriate. "
                                 f"Should be a flat list or a dict with None - values. Your job was: {data}")
            return data

        elif isinstance(data, dict):
            if all(v is None for v in data.values()):
                return list(data.keys())

            elif len(data.keys()) == 1:
                return [data]

            else:
                raise InvalidJob(f"Job have values with embedded data, but chunking is not requested. "
                                 f"You should either append 'isolate_ATTRIBUTE' flag or make values appropriate. "
                                 f"Should be a flat list or a dict with None - values. Your job was: {data}")

        else:
            raise InvalidJob(f"A Job without chunking enabled should have value either as a simple iterable "
                             f"(list, set, tuple), or a dict with all the values = None."
                             f"You provided {type(data)}: {data}")


    def last_x_days(self, pattern: str) -> List[str]:
        """
        Constructs the list of date strings for chunking.
        """

        assert re.match('last_[0-9]+_days', pattern) is not None, "Invalid pattern {pattern} for `last_x_days()`"

        num = int(pattern.split('_')[1])
        today = datetime.date.today()

        return [str(today - datetime.timedelta(days=x)) for x in range(num, 0, -1)]


    def previous_x_days(self, pattern: str) -> List[str]:
        """
        Returns a list of string dates from today - x - x

        For example, consider today's date as 2019-04-30.
        If I call for previous_x_days(pattern='previous_2_days'), I will receive a list of string dates equal to:
        ['2019-04-26', '2019-04-27']
        """
        assert re.match('previous_[0-9]+_days', pattern) is not None, "Invalid pattern {pattern} for `previous_x_days`"

        num = int(pattern.split('_')[1])
        today = datetime.date.today()
        end_date = today - datetime.timedelta(days=num)

        return [str(end_date - datetime.timedelta(days=x)) for x in range(num, 0, -1)]


    def x_days_back(self, pattern: str) -> List[str]:
        """
        Finds the exact date X days back from now.
        Returns it as a `str` in a `list` following the interface requirements of `chunk_dates`.

        e.g. `1_days_back` - yesterday, `7_days_back` - same day as today last week
        """

        assert re.match('[0-9]+_days_back', pattern) is not None, "Invalid pattern {pattern} for `x_days_back()`"

        num = int(pattern.split('_')[0])
        today = datetime.date.today()

        return [str(today - datetime.timedelta(days=num))]


    def yesterday(self, pattern: str = 'yesterday') -> List[str]:
        """
        Simple wrapper for x_days_back() to return yesterday's date.
        """
        assert re.match('yesterday', pattern) is not None, "Invalid pattern {pattern} for `yesterday()`"
        return self.x_days_back('1_days_back')


    def today(self, pattern: str = 'today') -> List[str]:
        """
        Returns list with one datetime string (YYYY-MM-DD) equal to today's date.
        """
        assert re.match('today', pattern) is not None, "Invalid pattern {pattern} for `today()`"
        return [str(datetime.date.today())]


    def last_week(self, pattern: str = 'last_week') -> List[str]:
        """
        Returns list of dates (YYYY-MM-DD) as strings for last week (Sunday - Saturday)
        :param pattern:
        :return:
        """
        assert re.match('last_week', pattern) is not None, "Invalid pattern {pattern} for `last_week()`"

        today = datetime.date.today()
        end_date = today - datetime.timedelta(days=today.weekday() + 8)

        return [str(end_date + datetime.timedelta(days=x)) for x in range(7)]


    def chunk_dates(self, job: Dict, skeleton: Dict = None) -> List[Dict]:
        """
        There is a support for multiple not nested parameters to chunk. Dates is one very specific of them.
        """

        data = []
        skeleton = deepcopy(skeleton) or {}
        job = deepcopy(job)

        period = job.pop('period', None)
        isolate = job.pop('isolate_days', None)

        PERIOD_KEYS = ['last_[0-9]+_days', '[0-9]+_days_back', 'yesterday', 'today', 'previous_[0-9]+_days',
                       'last_week']

        if period:

            date_list = []
            for pattern in PERIOD_KEYS:
                if re.match(pattern, period):
                    # Call the appropriate method with given value from job.
                    logger.debug(f"Found period '{period}' for job {job}")
                    method_name = pattern.replace('[0-9]+', 'x', 1)
                    date_list = getattr(self, method_name)(period)
                    break
            else:
                raise ValueError(f"Unsupported period requested: {period}. Valid options are: "
                                 f"'last_X_days', 'X_days_back', 'yesterday', 'today', 'previous_[0-9]+_days', 'last_week'")

            if isolate:
                assert len(date_list) > 0, f"The chunking period: {period} did not generate date_list. Bad."

                for d in date_list:
                    data.append({**job, **skeleton, 'date_list': [d]})
            else:
                if len(date_list) > 1:
                    logger.debug("Running chunking for multiple days, but without date isolation. "
                                 "Your workers might feel bad.")
                data.append({**job, **skeleton, 'date_list': date_list})

        else:
            logger.debug(f"No `period` chunking requested in job {job}")
            data.append({**job, **skeleton})

        return data


    def construct_job_data(self, job: Dict, skeleton: Dict = None) -> List[Dict]:
        """
        Chunks the job to tasks using several layers. Each layer is represented with a `chunker` method.
        All chunkers should accept `job` and optional `skeleton` for tasks and return a list of tasks.
        If there is nothing to chunk for some chunker, return same `job` (with injected `skeleton`) wrapped in a list.

        Default chunkers:

        - Date list chunking
        - Recursive chunking for `chunkable_attrs`

        """

        CHUNKERS = [self.chunk_dates, self.chunk_job]

        data = [job]
        skeleton = deepcopy(skeleton) or {}

        for chunker in CHUNKERS:
            chunked = []  # Container for results of current chunker method.
            for task in data:
                logging.debug(f"Chunking {task} with {chunker}")
                chunked.extend(chunker(job=task))

            data = deepcopy(chunked)

        # Inject the skeleton to the resulting tasks
        for task in data:
            task.update(skeleton)

        return data


    def chunk_job(self, job: dict, skeleton: Dict = None, attr: str = None) -> List[Dict]:
        """
        Recursively parses a job, validates everything and chunks to simple tasks what should be chunked.
        The Scenario of chunking and isolation is worth another story, so you should put a link here once it is ready.
        """

        data = []
        skeleton = deepcopy(skeleton) or {}
        job = deepcopy(job)

        # The current attribute we are looking for in this iteration or the first one of preconfigured chunkables.
        attr = attr or self.chunkable_attrs[0] if self.chunkable_attrs else None

        # We have to return here the full job to let it work correctly with recursive calls.
        if not attr:
            return [{**job, **skeleton}]

        logger.debug(f"Testing for chunking {attr} from {job} with skeleton {skeleton}")

        # First of all decide whether we need to chunk current job (or a sub-job if called recursively).
        if self.needs_chunking(plural(attr), job):

            # Next attribute is either name of attribute according to config, or None if we are already in last level.
            next_attr = self.get_next_chunkable_attr(attr)

            # Here and many places further we support both single and plural versions of attribute names.
            for possible_attr in single_or_plural(attr):
                current_vals = get_list_of_multiple_or_one_or_empty_from_dict(job, possible_attr)
                if not current_vals:
                    continue

                # This is not the `skeleton` received during the call, but the remaining parts of the `job`,
                # not related to current `attr`
                job_skeleton = {k: v for k, v in job.items() if k not in [possible_attr, f"isolate_{attr}s"]}
                logger.debug(f"For {possible_attr} we got current_vals: {current_vals} from {job}, "
                             f"leaving job_skeleton: {job_skeleton}")

                # For dictionaries we have to either go deeper recursively, or just flatten keys if values are None-s.
                if all(isinstance(v, dict) for v in current_vals):
                    for val in current_vals:
                        for name, subdata in val.items():
                            logger.debug(f"SubIterating `{name}` with {subdata}")

                            task = deepcopy(skeleton)
                            task.update(job_skeleton)
                            task[plural(attr)] = [name]

                            if isinstance(subdata, dict):
                                # print(f'DICT {subdata}, NEXTATTR {next_attr}')
                                if not next_attr:
                                    # If there is no lower level configured to chunk, just keep this subdata in payload
                                    task.update(subdata)
                                    data.append(task)
                                    # raise InvalidJob(f"Unexpected dictionary for unchunkable attribute: {attr}. "
                                    #                  f"In order to chunk this, you should support this level in: "
                                    #                  f"`config.job_schema.chunkable_attrs`. "
                                    #                  f"If you want to pass custom payload - put it as `payload` in "
                                    #                  f"your job. Job was: {job}")
                                else:
                                    logger.debug(f"Call recursive for {next_attr} from subdata: {subdata}")
                                    data.extend(self.chunk_job(job=subdata, skeleton=task, attr=next_attr))

                            # If None-s we just add a task. `Name` (which is actually a value in this scenario)
                            # was already added when creating task skeleton.
                            elif subdata is None:
                                logger.debug(f"Appending task to data for {name} from {val}")
                                data.append(task)

                            else:
                                raise InvalidJob(f"Unsupported type of val: {subdata} for attribute {possible_attr}")

                # If current vals are not dictionaries, we just validate that they are flat supported values
                else:
                    vals = self.validate_list_of_vals(current_vals)

                    # If batch_size is provided for this attr - we use it, by default chunk every.
                    batch_size = job.get(f'max_{plural(attr)}_per_batch', 1)

                    for val in chunks(vals, batch_size):
                        task = deepcopy(skeleton)
                        task.update(job_skeleton)
                        task[plural(attr)] = val
                        data.append(task)

        else:
            logger.debug(f"No need for chunking for attr: {attr} in job: {job}. Current skeleton is: {skeleton}")
            task = skeleton

            for a in single_or_plural(attr):
                if a in job:
                    attr_value = job.pop(a, None)
                    if attr_value:
                        try:
                            vals = self.validate_list_of_vals(attr_value)
                            task[plural(attr)] = vals
                        except InvalidJob:
                            logger.warning(f"Caught InvalidJob exception.")
                            # If a custom payload is not following the chunking convention - just translate it as is.
                            # And return the pop-ed value back to the job.
                            job[a] = attr_value
                        break
            else:
                logger.error(f"Did not find values for {attr} in job: {job}")
            # Populate the remaining parts of the job back to task.
            task.update(job)

            logger.debug(f"Appending task to data: {task}")
            data.append(task)

        return data


    @staticmethod
    def get_index_from_list(attr, data):
        """ Finds the index ignoring the 's' at the end of attribute. """

        assert isinstance(data, Iterable), f"Non iterable data for get_index_from_list: {data}"
        assert isinstance(attr, str), f"Non-string attr for get_index_from_list: {type(attr)}"

        attrs = single_or_plural(attr)
        for a in attrs:
            try:
                return list(data).index(a)
            except ValueError:
                pass
        raise ValueError(f"Not found {attr} in {data}")


    def get_next_chunkable_attr(self, attr):
        """ Return the next by order after `attr` chunkable attribute. """

        attrs = single_or_plural(attr)
        for a in attrs:
            try:
                return self.chunkable_attrs[self.get_index_from_list(a, self.chunkable_attrs) + 1]
            except (IndexError, KeyError, TypeError, ValueError):
                pass


    def needs_chunking(self, attr: str, data: Dict) -> bool:
        """
        Recursively analyses the data and identifies if the current level of data should be chunked.
        This could happen if either isolate_attr marker in the current scope or recursively in any of sub-elements.

        :param attr:    Name of attribute you want to check for chunking.
        :param data:    Input dictionary to analyse.
        """

        attrs = single_or_plural(attr)
        isolate_attrs = [f"isolate_{a}" for a in attrs] + [f"max_{a}_per_batch" for a in attrs]

        if any(data[x] for x in isolate_attrs if x in data):
            logger.debug(f"needs_chunking(): Got requirement to isolate {attr} in the current scope: {data}")
            return True

        next_attr = self.get_next_chunkable_attr(attr)

        logger.debug(f"needs_chunking(): Found next attr {next_attr}, for {attr} from {data}")
        # We are not yet lowest level going recursive
        if next_attr:
            for a in attrs:
                current_vals = get_list_of_multiple_or_one_or_empty_from_dict(data, a)
                logger.debug(f"needs_chunking(): For {a} got current_vals: {current_vals} from {data}. "
                             f"Analysing {next_attr}")

                for val in current_vals:

                    for name, subdata in val.items():
                        logger.debug(f"needs_chunking(): Analysing {next_attr} in {subdata}")
                        if not subdata:
                            continue
                        logger.debug(f"needs_chunking(): Going recursive for {next_attr} in {subdata}")
                        if self.needs_chunking(next_attr, subdata):
                            logger.debug(f"needs_chunking(): Returning True for {next_attr} from {subdata}")
                            return True

        return False


    def extract_job_from_payload(self, event: Dict):
        """ Parse and basically validate job from the event. """


        def load(obj):
            return obj if isinstance(obj, dict) else json.loads(obj)


        jh = load(event)
        job = load(jh['job']) if 'job' in jh else jh

        assert 'lambda_name' in job, f"Job is missing required parameter 'lambda_name': {job}"
        job['lambda_name'] = trim_arn_to_name(job['lambda_name'])

        return job


    def process_file(self):
        """
        Process a file for creating tasks, then uploading it to S3.
        In case of execution time reached its limit, spawning a new sibling to continue the processing.

        """

        file_name = self.get_and_lock_queue_file()

        if not file_name:
            logger.info(f"No file in queue.")
            return

        else:
            while self.sufficient_execution_time_left:
                data = self.pop_rows_from_file(file_name, rows=self._rows_to_process)
                if not data:
                    break

                for task in data:
                    logger.info(task)
                    t = json.loads(task)
                    labourer = self.task_client.get_labourer(t['labourer_id'])
                    self.task_client.create_task(labourer=labourer, **t)
                    time.sleep(self._sleeptime_for_dynamo)

            else:
                # Spawning another sibling to continue the processing
                try:
                    payload = dict(file_name=file_name)
                    self.siblings_client.spawn_sibling(global_vars.lambda_context, payload=payload)
                    self.stats['siblings_spawned'] += 1

                except Exception as err:
                    logger.exception(f"Could not spawn sibling with context: {global_vars.lambda_context}, payload: {payload}")

            self.upload_and_unlock_queue_file()
            self.clean_tmp()


    @property
    def _sleeptime_for_dynamo(self):
        """
        Pull DynamoDB write capcity dynamically to throttle at appropriate levels

        Calculates based on the assumption that a single write action consumes a full WCU
        Therefore multiple capacity units are calculated as a fraction of the
        """
        logging.debug(dir(self.task_client.dynamo_db_client))
        return 1 / self.task_client.dynamo_db_client.get_capacity()['write']


    @staticmethod
    def pop_rows_from_file(file_name: str, rows: Optional[int] = 1) -> List[str]:
        """
        Reads the rows from the top of file. Along the way removes them from original file.

        :param str file_name:    File to read.
        :param int rows:        Number of rows to read. Default: 1
        :return:                List of strings read from file top.
        """

        tmp_file = f"/tmp/in_prog_{file_name.replace('/', '_')}"
        result = []

        try:
            with open(file_name) as f, open(tmp_file, "w") as out:
                for _ in range(rows):
                    try:
                        result.append(next(f))
                    except StopIteration:
                        break

                # Writing remaining rows to the temp file.
                for line in f:
                    out.write(line)

            os.remove(file_name)
            os.rename(tmp_file, file_name)

            # If there is no dat remaining in the file we remove it.
            if os.path.getsize(file_name) == 0:
                os.remove(file_name)

        except FileNotFoundError:
            pass

        return result


    def clean_tmp(self, file_name=None):
        file_to_remove = file_name or self.local_queue_file

        if os.path.isfile(file_to_remove):
            os.remove(file_to_remove)


    @property
    def _rows_to_process(self):
        return self.config['rows_to_process']


    @property
    def sufficient_execution_time_left(self) -> bool:
        """
        Return if there is a sufficient execution time for processing ('shutdown period' is in seconds).
        """

        return global_vars.lambda_context.get_remaining_time_in_millis() > self.config['shutdown_period'] * 1000


    def get_and_lock_queue_file(self) -> str:
        """
        Either take a new (recently created) file in local /tmp/, or download the version of queue file from S3.
        We move the file in S3 to `locked_` by prefix state or simply upload the new one there in `locked_` state.

        :return: Local path to the file.
        """

        if not os.path.isfile(self.local_queue_file):
            try:
                self.s3_client.download_file(Bucket=self._queue_bucket, Key=self.remote_queue_file,
                                             Filename=self.local_queue_file)
            except self.s3_client.exceptions.ClientError:
                self.stats['non_existing_remote_queue'] += 1
                logger.exception(f"Not found remote file to download")

            else:
                self.s3_client.copy_object(Bucket=self._queue_bucket,
                                           CopySource=f"{self._queue_bucket}/{self.remote_queue_file}",
                                           Key=self.remote_queue_locked_file)

                self.s3_client.delete_object(Bucket=self._queue_bucket, Key=self.remote_queue_file)

                logger.debug(f"Downloaded a copy of {self.local_queue_file} for processing "
                             f"and moved the remote one to {self.remote_queue_locked_file}.")

        # If the local file exists (means we have probably just created it). Then we upload it in `locked_` state.
        else:
            self.s3_client.upload_file(Filename=self.local_queue_file, Bucket=self._queue_bucket,
                                       Key=self.remote_queue_locked_file)

        return self.local_queue_file


    def upload_and_unlock_queue_file(self):
        """
        Upload the local queue file to S3 and remove the `locked_` by prefix copy if it exists.
        """

        # If there is data left unprocessed in the file, upload it for future processing by siblings or someone else.
        if os.path.isfile(self.local_queue_file):
            self.s3_client.upload_file(Filename=self.local_queue_file, Bucket=self._queue_bucket,
                                       Key=self.remote_queue_file)

        # Delete the locked file from S3 (aka unlock)
        try:
            self.s3_client.delete_object(Bucket=self._queue_bucket, Key=self.remote_queue_locked_file)
        except self.s3_client.exceptions.ClientError:
            logger.debug(f"No remote locked file to remove: {self.remote_queue_locked_file}. This is probably new.")


    @property
    def _queue_bucket(self):
        """ Name of S3 bucket for file with queue of tasks not yet in DynamoDB. """
        return self.config['queue_bucket']


    def set_queue_file(self, name: str = None):
        """
        Initialize a unique file_name to store the queue of tasks to write.
        """

        if name is None:
            filename_parts = self.config['queue_file'].rsplit('.', 1)
            assert len(filename_parts) == 2, "Got bad file name"
            self._queue_file_name = f"{filename_parts[0]}_{global_vars.lambda_context.aws_request_id}.{filename_parts[1]}"
        else:
            self._queue_file_name = name


    @property
    def local_queue_file(self):
        return f"/tmp/{self._queue_file_name}"


    @property
    def remote_queue_file(self):
        """ Full S3 Key of file with queue of tasks not yet in DynamoDB. """
        return f"{self.config['s3_prefix'].strip('/')}/{self._queue_file_name}"


    @property
    def remote_queue_locked_file(self):
        """
        Full S3 Key of file with queue of tasks not yet in DynamoDB in the `locked` state.
        Concurrent processes should not touch it.
        """
        return f"{self.config['s3_prefix'].strip('/')}/locked_{self._queue_file_name}"

global_vars = LambdaGlobals()



import boto3
import csv
import json
import logging
import os
from collections import defaultdict


__author__ = "Nikolay Grishchenko"
__email__ = "dev@bimpression.com"
__version__ = "1.00"
__license__ = "MIT"
__status__ = "Production"

__all__ = ['SnsManager']

logger = logging.getLogger()


class SnsManager():
    """
    AWS Simple Notification System Manager helper.
    Requires the Role to have permissions to access SSM for requested resources, something like.

    Must have a recipient specified either during initialization or later using set_recipient method.
    Messages received by send_message() are batched and will be actually send to SNS only during the
    call of commit() or during the destruction of the class.

    The method doesn't yet support batching messages for multiple recipients, so in case you try to
    change the recipient it automatically sends all the currently batched messages to the previous recipient.
    """


    def __init__(self, **kwargs):
        """
        :param recipient:   - str   - ARN of the default recipient.
        :param subject:     - str   - Default subject for messages.
        """

        self.stats = defaultdict(int)
        config = kwargs.get('config', kwargs.get('custom_config', {}))
        if config:
            self.recipient = config.get('recipient')
            self.subject = config.get('subject')
        else:
            self.recipient = kwargs.get('recipient')
            self.subject = kwargs.get('subject')

        self.queue = []

        self.test = kwargs.get('test') or True if os.environ.get('STAGE') == 'test' else False
        if self.test:
            self.recipient = 'arn:aws:sns:us-west-2:000000000000:autotest_topic'

        if not self.test:
            self.session = boto3.Session(region_name=kwargs.get('region', 'us-west-2'))
            self.resource = self.session.client('sns')


    def __del__(self):
        """
        Destructor. Send unsent queued messages.
        """

        if self.queue:
            self.commit()


    def set_client_attr(self, name, value):
        """
        Sets the given _name_ attribute to _value_.
        Commits the self.queue if there are any messages in it.
        """

        if getattr(self, name) != value:
            if self.queue:
                self.commit()

            setattr(self, name, value)


    def set_recipient(self, arn):
        assert isinstance(arn, str), f"Invalid format of ARN: {arn}. Recipient must be string ARN of SNS Topic"
        assert arn.lower().startswith('arn:aws:'), f"Invalid format of ARN: {arn}. Recipient must be ARN of SNS Topic"
        self.set_client_attr('recipient', value=arn)


    def set_subject(self, value):
        self.set_client_attr('subject', value=str(value))


    def commit(self):
        """
        Combines messages from self.queue and pushes them to self.recipient.
        Cleans the queue after that.
        """

        # Check that the ARN of recipient is set.
        if not self.recipient:
            raise RuntimeError("You did not specify ARN of recipient to send message to. "
                               "Please use self.sns.set_recipient() from your Lambda __")

        # Check that the SNS subject is set.
        if not self.subject:
            raise RuntimeError("You did not specify Subject for the message. "
                               "We don't want you to write code like this, please fix.")

        message = "\n\n#####\n\n".join(self.queue)

        if message:
            self.resource.publish(
                    TopicArn=self.recipient,
                    Subject=self.subject,
                    Message=message)

        self.queue = []


    def send_message(self, message, subject=None, forse_commit=False):
        """
        If the subject is not yet set (for example during __init__() of the class) - then require subject to be set.
        Otherwize we accept None subject and simply append messages to queue.
        Once the subject changes - the queue is commite automatically.

        :param message:     - str   - Message to be send in body of SNS message. Queued.
        :param subject:     - str   - Optional. Custom subject for message.
        """

        if not any([self.subject, subject]):
            raise RuntimeError("You must have specified subject for self.sns.send_message() either "
                               "during __init__() of the class, or in the first send_message in kwargs.")

        if all([self.subject, subject]) and not self.subject == subject:
            logger.info("Change of subject detected. We commit (send) the current queue.")
            self.set_subject(subject)  # This will also commit existing messages automatically.
            self.queue = [message]
        else:
            self.queue.append(message)

        if forse_commit:
            if subject and not self.subject:
                self.subject = subject
            logger.info("The caller asked to forse_commit, so we commit the queue immediately.")
            self.commit()



# import logging
# import unittest
# import os
# import boto3
#
# from unittest.mock import patch, MagicMock
#
# from sosw.components.tasks_api_client_for_workers import *
#
#
# logger = logging.getLogger()
#
# os.environ["STAGE"] = "test"
# os.environ["autotest"] = "True"
#
# TASK_ID = 'test_task_id'
# LABOURER_ID = 'some_lambda'
# TABLE_NAME = 'autotest_sosw_tasks'
#
# TEST_ITEM = {
#     'task_id':     TASK_ID,
#     'labourer_id': LABOURER_ID
# }
#
# TEST_ITEM_DICT = {
#     'task_id':     {'S': TASK_ID}
# }
#
#
# class tasks_api_client_for_workers_TestCase(unittest.TestCase):
#
#     def setUp(self):
#         """ Setting up an item in dynamo DB """
#
#         self.client = boto3.client('dynamodb')
#         response = self.client.put_item(TableName=TABLE_NAME, Item=TEST_ITEM_DICT)
#         if response['ResponseMetadata']['HTTPStatusCode'] == 200:
#             logger.info("Item created successfully.")
#         else:
#             raise RuntimeError("Item creation has failed.")
#
#
#     def tearDown(self):
#         """ Cleaning up the created item from dynamo DB """
#
#         response = self.client.delete_item(TableName=TABLE_NAME, Key=TEST_ITEM_DICT)
#         if response['ResponseMetadata']['HTTPStatusCode'] == 200:
#             logger.info("Item deleted successfully.")
#         else:
#             raise RuntimeError("Item deletion has failed.")
#
#
#     def test_close_task(self):
#         """ Testing the 'close_task()' function """
#         with patch('time.time') as t:
#             t.return_value = 9500
#             close_task(TEST_ITEM)
#
#         result = self.client.get_item(TableName=TABLE_NAME, Key=TEST_ITEM_DICT)['Item']
#         self.assertEqual(result['task_id'], {'S': TASK_ID})
#         self.assertEqual(result['completed_at'], {'N': '9500'})
#
#
# if __name__ == '__main__':
#     unittest.main()



import boto3
import logging
import shutil
import unittest
import uuid
import os
import csv

from collections import defaultdict
from unittest.mock import MagicMock
from sosw.components.sns import *


logging.getLogger('botocore').setLevel(logging.WARNING)

os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"


class sns_TestCase(unittest.TestCase):

    def clean_queue(self):
        setattr(self.sns, 'queue', [])


    def setUp(self):
        self.sns = SnsManager(test=True, subject='Autotest SNS Subject')
        self.sns.commit = MagicMock(side_effect=self.clean_queue)

    def tearDown(self):
        pass


    def test_init__reads_config(self):

        sns = SnsManager(config={'subject': 'subj', 'recepient': 'arn::some_topic'})

        self.assertEqual(sns.recipient, 'arn:aws:sns:us-west-2:000000000000:autotest_topic',
                         "The Topic must be automatically reset for test")
        self.assertEqual(sns.subject, 'subj', "Subject was not set during __init__ from config.")


    def test_queue_message(self):
        self.sns.send_message("test message")
        self.assertEqual(len(self.sns.queue), 1, "Default send_message() did not queue the message.")


    def test_queue_message_with_subject(self):
        self.sns.send_message("test message", subject="New Subject")
        self.assertEqual(len(self.sns.queue), 1, "send_message() with custom subject did not queue.")


    def test_commit_queue(self):
        self.sns.send_message("test message")
        self.sns.commit()
        self.assertEqual(len(self.sns.queue), 0, f"Commit did not clean queue")
        self.sns.commit.assert_called_once()


    def test_commit_on_change_subject(self):
        self.sns.send_message("test message")
        self.sns.set_subject("New Subject")
        self.assertEqual(len(self.sns.queue), 0, "On change subject the queue should be committed.")


    def test_no_commit_on_change_subject_if_subject_is_same(self):
        self.sns.send_message("test message")
        self.sns.set_subject("Autotest SNS Subject")
        self.assertEqual(len(self.sns.queue), 1, "On change subject the queue should be committed.")


    def test_no_commit_on_same_subject(self):
        self.sns.send_message("test message")
        self.sns.send_message("test message", subject="Autotest SNS Subject")
        self.assertEqual(len(self.sns.queue), 2, "On sending message with exactly same subject, it should be queued.")


    def test_commit_and_queue_on_change_subject(self):
        self.sns.send_message("test message")
        self.assertEqual(len(self.sns.queue), 1)
        self.sns.send_message("test message", subject="New Subject")
        self.assertEqual(len(self.sns.queue), 1, "On change subject, old message should be committed, new one queued.")


    def test_commit_auto_on_change_recipient(self):
        self.sns.send_message("test message")
        self.assertEqual(len(self.sns.queue), 1, f"Initial send_message() did not queue the message")

        self.sns.set_recipient('arn:aws:sns:new_recipient')
        self.assertEqual(len(self.sns.queue), 0)


    def test_no_commit_on_change_recipient_if_recipient_is_same(self):
        self.sns.send_message("test message")
        self.assertEqual(len(self.sns.queue), 1, f"Initial send_message() did not queue the message")

        self.sns.set_recipient('arn:aws:sns:us-west-2:000000000000:autotest_topic')
        self.assertEqual(len(self.sns.queue), 1)


    def test_validate_recipient(self):
        """
        Must be a string with ARN of SNS Topic. Validator just checks that string starts with 'arn:aws:'
        """
        self.assertRaises(AssertionError, self.sns.set_recipient, 'just_new_recipient_not_full_arn')


    # def test_true(self):
    #     self.assertTrue(False, "Break the test not to deploy")


if __name__ == '__main__':
    unittest.main()



__all__ = ['EcologyManager', 'ECO_STATUSES']
__author__ = "Nikolay Grishchenko"
__version__ = "1.0"

import boto3
import datetime
import json
import logging
import operator
import os
import random
import time

from collections import defaultdict
from collections import OrderedDict
from statistics import mean
from typing import Dict, List, Optional, Union

from sosw.app import Processor
from sosw.labourer import Labourer
from sosw.components.benchmark import benchmark
from sosw.components.helpers import make_hash
from sosw.managers.task import TaskManager


logger = logging.getLogger()
logger.setLevel(logging.INFO)

ECO_STATUSES = (
    (0, 'Bad'),
    (1, 'Poor'),
    (2, 'Moderate'),
    (3, 'Good'),
    (4, 'High'),
)


class EcologyManager(Processor):
    DEFAULT_CONFIG = {
        'init_clients':               ['cloudwatch'],
        'default_metric_values':
                                      {
                                          'Period':     60,
                                          'Statistics': ['Average'],
                                          'MetricAggregationTimeSlice': 300
                                      }
    }

    running_tasks = defaultdict(int)
    health_metrics: Dict = None
    task_client: TaskManager = None  # Will be Circular import! Careful!
    cloudwatch_client: boto3.client = None


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def __call__(self, event):
        raise NotImplementedError


    def register_task_manager(self, task_manager: TaskManager):
        """
        We will have to make some queries, and don't want to initialise another TaskManager locally.
        Just receive the pointer to TaskManager from whoever needs.

        This could be in __init__, but I don't want to update the initialization workflows for every function
        initialising me. They usually use built-in in core Processor mechanism to register_clients().
        """

        logger.info("Registering TaskManager for EcologyManager")
        self.task_client = task_manager

        logger.info("Reset cache of running_tasks counter in EcologyManager")
        self.running_tasks = defaultdict(int)

        logger.info("Reset cache of health_metrics in EcologyManager")
        self.health_metrics = dict()


    @property
    def eco_statuses(self):
        return [x[0] for x in ECO_STATUSES]


    def fetch_metric_stats(self, metric: Dict) -> List[Dict]:
        """
        Fetches from CloudWatch Datapoints of aggregated metric statistics.
        Fields in `metric` are the attributes of get_metric_statistics_.
        Additional parameter: MetricAggregationTimeSlice in seconds is used to calculate the Start and EndTime.

        .. _get_metric_statistics: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudwatch.html#CloudWatch.Client.get_metric_statistics

        If some fields are missing in the metric, the defaults come from ``config['default_metric_values']``
        """

        _cfg = self.config.get

        COMPARATORS = {
            'Average': mean,
            'Maximum': max,
            'Minimum': min,
        }

        params = metric.copy()

        # Setting up default metric parameters if unspecified for current metric.
        for k, v in _cfg('default_metric_values', {}).items():
            if k not in metric:
                params[k] = v
                logger.debug(f"Set the default {v} for {k} in metric query {metric}.")

        assert len(params['Statistics']) == 1, "Complex statistics aggregation is not yet supported"

        duration = int(params.pop('MetricAggregationTimeSlice'))

        params['EndTime'] = datetime.datetime.now()
        params['StartTime'] = params['EndTime'] - datetime.timedelta(seconds=duration)

        logger.debug(f"Query to CloudWatch `get_metric_statistics`: {params}")
        result = self.cloudwatch_client.get_metric_statistics(**params)

        comparator_name = params['Statistics'][0]
        comparator = COMPARATORS[comparator_name]

        return comparator(x[comparator_name] for x in result.get('Datapoints', list()))


    def get_labourer_status(self, labourer: Labourer) -> int:
        """
        Get the worst (lowest) health status according to preconfigured health metrics of the Labourer.

        .. _ECO_STATUSES:

        Current ECO_STATUSES:

        - (0, 'Bad')
        - (1, 'Poor')
        - (2, 'Moderate')
        - (3, 'Good')
        - (4, 'High')
        """

        _cfg = self.config.get

        health = max(map(lambda x: int(x[0]), ECO_STATUSES))

        metrics = getattr(labourer, 'health_metrics', {}) or {}
        for health_metric in metrics.values():

            metric_hash = make_hash(health_metric['details'])
            if metric_hash not in self.health_metrics:
                self.health_metrics[metric_hash] = self.fetch_metric_stats(metric=health_metric['details'])
                logger.info(f"Updated the cache of Ecology metric {metric_hash} - {health_metric} "
                            f"with {self.health_metrics[metric_hash]}")

            value = self.health_metrics[metric_hash]
            logger.debug(f"Ecology metric {metric_hash} has {value}")

            health = min(health, self.get_health(value, metric=health_metric))

        logger.info(f"Ecology health of Labourer {labourer} is {health}")

        return health


    def get_health(self, value: Union[int, float], metric: Dict) -> int:
        """
        Checks the value against the health_metric configuration.
        """

        op = getattr(operator, metric.get('feeling_comparison_operator'))

        # Find the first configured feeling from the map that does not comply.
        # Order and validate the feelings
        feelings = OrderedDict([(key, metric['feelings'][key])
                                for key in sorted(metric['feelings'].keys(), reverse=True)])

        last_target = 0
        for health, target in feelings.items():
            if op(target, last_target):
                raise ValueError(f"Order of values if feelings is invalid and doesn't match expected eco statuses: "
                                 f"{feelings.items()}. Failed: {last_target} not "
                                 f"{metric.get('feeling_comparison_operator')} {target}")

            if op(value, target):
                return int(health)

            last_target = target

        return 0


    def count_running_tasks_for_labourer(self, labourer: Labourer) -> int:
        """
        TODO Refactor this to cache the value in the Labourer object itself.
        Should also update add_running_tasks_for_labourer() for that.
        """

        if not self.task_client:
            raise RuntimeError("EcologyManager doesn't have a TaskManager registered. "
                               "You have to call register_task_manager() after initiazation and pass the pointer "
                               "to your TaskManager instance.")

        if labourer.id not in self.running_tasks.keys():
            self.running_tasks[labourer.id] = self.task_client.get_count_of_running_tasks_for_labourer(labourer)
            logger.debug(f"EcologyManager.count_running_tasks_for_labourer() recalculated cache for Labourer "
                         f"{labourer}")

        logger.debug(f"EcologyManager.count_running_tasks_for_labourer() returns: {self.running_tasks[labourer.id]}")
        return self.running_tasks[labourer.id]


    def add_running_tasks_for_labourer(self, labourer: Labourer, count: int = 1):
        """
        Adds to the current counter of running tasks the given `count`.
        Invokes the getter first in case the original number was not yet calculated from DynamoDB.
        """

        self.running_tasks[labourer.id] = self.count_running_tasks_for_labourer(labourer) + count


    def get_labourer_average_duration(self, labourer: Labourer) -> int:
        """
        Calculates the average duration of `labourer` executions.

        The operation consumes DynamoDB RCU . Normally this method is called for each labourer only once during
        registration of Labourers. If you want to learn this value, you should ask Labourer object.

        .. code-block::python

           some_labourer.get_attr('average_duration')
        """

        if not self.task_client:
            raise RuntimeError("EcologyManager doesn't have a TaskManager registered. "
                               "You have to call register_task_manager() after initiazation and pass the pointer "
                               "to your TaskManager instance.")

        return self.task_client.get_average_labourer_duration(labourer)


    def get_max_labourer_duration(self, labourer: Labourer) -> int:
        """
        Maximum duration of `labourer` executions.
        """

        resp = self.task_client.lambda_client.get_function_configuration(FunctionName=labourer.arn)
        return resp['Timeout']


    # The task_client of EcologyManager is just a pointer. We skip recursive stats to avoid infinite loop.
    def get_stats(self, recursive=False):
        return super().get_stats(recursive=False)


    def reset_stats(self, recursive=False):
        return super().reset_stats(recursive=False)



import os
import unittest

from unittest.mock import patch, MagicMock

from sosw.worker import Worker


os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"


class Worker_UnitTestCase(unittest.TestCase):


    def setUp(self):
        self.patcher = patch("sosw.app.get_config")
        self.get_config_patch = self.patcher.start()


    def tearDown(self):
        self.patcher.stop()

        try:
            del (os.environ['AWS_LAMBDA_FUNCTION_NAME'])
        except Exception:
            pass


    def test_close_task__called(self):
        with patch('boto3.client'):
            p = Worker()

        p.mark_task_as_completed = MagicMock(return_value=None)

        p({'task_id': '123'})
        p.mark_task_as_completed.assert_called_once_with('123')



import logging
import unittest
import os

from unittest.mock import MagicMock, patch, Mock


logging.getLogger('botocore').setLevel(logging.WARNING)

os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"

from sosw.components.dynamo_db import DynamoDbClient


class dynamodb_client_UnitTestCase(unittest.TestCase):
    TEST_CONFIG = {
        'row_mapper':      {
            'lambda_name':   'S',
            'invocation_id': 'S',
            'en_time':       'N',

            'hash_col':      'S',
            'range_col':     'N',
            'other_col':     'S',
            'new_col':       'S',
            'some_col':      'S',
            'some_counter':  'N',
            'some_bool':     'BOOL',
            'some_bool2':    'BOOL',
        },
        'required_fields': ['lambda_name'],
        'table_name':      'autotest_dynamo_db'
    }


    def setUp(self):
        self.HASH_KEY = ('hash_col', 'S')
        self.RANGE_KEY = ('range_col', 'N')
        self.KEYS = ('hash_col', 'range_col')
        self.table_name = 'autotest_dynamo_db'

        self.patcher = patch("boto3.client")
        self.paginator_mock = MagicMock()
        self.dynamo_mock = MagicMock()
        self.dynamo_mock.get_paginator.return_value = self.paginator_mock

        self.boto3_client_patch = self.patcher.start()
        self.boto3_client_patch.return_value = self.dynamo_mock

        self.dynamo_client = DynamoDbClient(config=self.TEST_CONFIG)


    def tearDown(self):
        self.patcher.stop()


    def test_dict_to_dynamo_strict(self):
        dict_row = {'lambda_name': 'test_name', 'invocation_id': 'test_id', 'en_time': 123456, 'some_bool': True,
                    'some_bool2': 'True'}
        dynamo_row = self.dynamo_client.dict_to_dynamo(dict_row)
        expected = {
            'lambda_name': {'S': 'test_name'}, 'invocation_id': {'S': 'test_id'}, 'en_time': {'N': '123456'},
            'some_bool': {'BOOL': True}, 'some_bool2': {'BOOL': True}
        }
        for key in expected.keys():
            self.assertDictEqual(expected[key], dynamo_row[key])


    def test_dict_to_dynamo_not_strict(self):
        dict_row = {'name': 'cat', 'age': 3, 'other_bool': False, 'other_bool2': 'False'}
        dynamo_row = self.dynamo_client.dict_to_dynamo(dict_row, strict=False)
        expected = {'name': {'S': 'cat'}, 'age': {'N': '3'}, 'other_bool': {'BOOL': False}}
        for key in expected.keys():
            self.assertDictEqual(expected[key], dynamo_row[key])


    def test_dict_to_dynamo_prefix(self):
        dict_row = {'hash_col': 'cat', 'range_col': '123', 'some_col': 'no'}
        dynamo_row = self.dynamo_client.dict_to_dynamo(dict_row, add_prefix="#")
        expected = {'#hash_col': {'S': 'cat'}, '#range_col': {'N': '123'}, '#some_col': {'S': 'no'}}
        for key in expected.keys():
            self.assertDictEqual(expected[key], dynamo_row[key])


    def test_dynamo_to_dict(self):
        dynamo_row = {
            'lambda_name': {'S': 'test_name'}, 'invocation_id': {'S': 'test_id'}, 'en_time': {'N': '123456'},
            'extra_key':   {'N': '42'}, 'some_bool': {'BOOL': False}
        }
        dict_row = self.dynamo_client.dynamo_to_dict(dynamo_row)
        expected = {'lambda_name': 'test_name', 'invocation_id': 'test_id', 'en_time': 123456, 'some_bool': False}
        self.assertDictEqual(dict_row, expected)


    def test_dynamo_to_dict_no_strict_row_mapper(self):
        dynamo_row = {
            'lambda_name': {'S': 'test_name'}, 'invocation_id': {'S': 'test_id'}, 'en_time': {'N': '123456'},
            'extra_key_n': {'N': '42'}, 'extra_key_s': {'S': 'wowie'}, 'other_bool': {'BOOL': True}
        }
        dict_row = self.dynamo_client.dynamo_to_dict(dynamo_row, fetch_all_fields=True)
        expected = {
            'lambda_name': 'test_name', 'invocation_id': 'test_id', 'en_time': 123456, 'extra_key_n': 42,
            'extra_key_s': 'wowie', 'other_bool': True
        }
        self.assertDictEqual(dict_row, expected)


    def test_dynamo_to_dict__dont_json_loads(self):
        config = self.TEST_CONFIG.copy()
        config['dont_json_loads_results'] = True

        self.dynamo_client = DynamoDbClient(config=config)

        dynamo_row = {
            'hash_col':   {'S': 'aaa'}, 'range_col': {'N': '123'}, 'other_col': {'S': '{"how many": 300}'},
            'duck_quack': {'S': '{"quack": "duck"}'}
        }
        res = self.dynamo_client.dynamo_to_dict(dynamo_row, fetch_all_fields=True)
        expected = {
            'hash_col': 'aaa', 'range_col': 123, 'other_col': '{"how many": 300}', 'duck_quack': '{"quack": "duck"}'
        }
        self.assertDictEqual(res, expected)

        res = self.dynamo_client.dynamo_to_dict(dynamo_row, fetch_all_fields=False)
        expected = {
            'hash_col': 'aaa', 'range_col': 123, 'other_col': '{"how many": 300}'
        }
        self.assertDictEqual(res, expected)


    def test_dynamo_to_dict__do_json_loads(self):
        config = self.TEST_CONFIG.copy()
        config['dont_json_loads_results'] = False

        self.dynamo_client = DynamoDbClient(config=config)

        dynamo_row = {
            'hash_col':   {'S': 'aaa'}, 'range_col': {'N': '123'}, 'other_col': {'S': '{"how many": 300}'},
            'duck_quack': {'S': '{"quack": "duck"}'}
        }
        res = self.dynamo_client.dynamo_to_dict(dynamo_row, fetch_all_fields=True)
        expected = {
            'hash_col': 'aaa', 'range_col': 123, 'other_col': {"how many": 300}, 'duck_quack': {"quack": "duck"}
        }
        self.assertDictEqual(res, expected)

        res = self.dynamo_client.dynamo_to_dict(dynamo_row, fetch_all_fields=False)
        expected = {
            'hash_col': 'aaa', 'range_col': 123, 'other_col': {"how many": 300}
        }
        self.assertDictEqual(res, expected)



    def test_get_by_query__validates_comparison(self):
        self.assertRaises(AssertionError, self.dynamo_client.get_by_query, keys={'k': '1'},
                          comparisons={'k': 'unsupported'})


    def test_get_by_query__between(self):
        keys = {'hash_col': 'cat', 'st_between_range_col': '3', 'en_between_range_col': '6'}

        self.dynamo_client = DynamoDbClient(config=self.TEST_CONFIG)

        self.dynamo_client.get_by_query(keys=keys)
        # print(f"Call_args for paginate: {self.paginator_mock.paginate.call_args}")

        args, kwargs = self.paginator_mock.paginate.call_args
        # print(kwargs)

        self.assertEqual(len(kwargs['ExpressionAttributeValues']), 3)
        self.assertIn('range_col between :st_between_range_col and :en_between_range_col',
                      kwargs['KeyConditionExpression'])


    def test_get_by_query__return_count(self):

        # Make sure dynamo paginator is mocked.
        self.paginator_mock.paginate.return_value = [{'Count': 24, 'LastEvaluatedKey': 'bzz'}, {'Count': 12}]
        self.dynamo_client.dynamo_client.get_paginator.return_value = self.paginator_mock

        # Call the manager
        result = self.dynamo_client.get_by_query(keys={'a': 'b'}, return_count=True)

        # Validate result
        self.assertEqual(result, 36, f"Result from 2 pages should be 24 + 12, but we received: {result}")

        # Make sure the paginator was called
        self.dynamo_client.dynamo_client.get_paginator.assert_called()


    def test__parse_filter_expression(self):
        TESTS = {
            'key = 42': ("key = :filter_key", {":filter_key": {'N': '42'}}),
            '   key    = 42  ': ("key = :filter_key", {":filter_key": {'N': '42'}}),
            'cat = meaw': ("cat = :filter_cat", {":filter_cat": {'S': 'meaw'}}),
            'magic between 41 and 42': ("magic between :st_between_magic and :en_between_magic",
                                        {":st_between_magic": {'N': '41'}, ":en_between_magic": {'N': '42'}}),
            'attribute_not_exists boo': ("attribute_not_exists (boo)", {})
        }

        for data, expected in TESTS.items():
            self.assertEqual(self.dynamo_client._parse_filter_expression(data), expected)


    def test__parse_filter_expression__raises(self):

        TESTS = [
            {'k': 1}, [1,2], None,  # Invalid input types
            'key == 42', 'foo ~ 1', 'foo3 <=> 0', 'key between 42',  # Invalid operators
            'key between 23, 25', 'key between [23, 25]', 'key 23 between 21',  # Invalid between formats.
        ]

        for data in TESTS:
            self.assertRaises((AssertionError, ValueError), self.dynamo_client._parse_filter_expression, data)


    def test_create__calls_put(self):
        row = {'hash_col': 'cat', 'range_key': 'test', 'another_col': 'wow'}
        self.dynamo_client.put = MagicMock(return_value=None)

        self.dynamo_client.create(row)

        self.dynamo_client.put.assert_called_once_with(row, None, overwrite_existing=False)


    def test_batch_get_items_one_table__strict(self):
        # Strict - returns only fields that are in the row mapper
        db_items = [{'hash_col': {'S': 'b'}, 'range_col': {'N': '10'}, 'unknown_col': {'S': 'not_strict'}}]
        db_result = {'Responses': {'autotest_dynamo_db': db_items}}

        self.dynamo_client.dynamo_client.batch_get_item = Mock(return_value=db_result)

        result = self.dynamo_client.batch_get_items_one_table(keys_list=[{'hash_col': 'b'}], fetch_all_fields=False)

        self.assertEqual(result, [{'hash_col': 'b', 'range_col': 10}])


    def test_batch_get_items_one_table__not_strict(self):
        # Not strict - returns all fields
        db_items = [{'hash_col': {'S': 'b'}, 'range_col': {'N': '10'}, 'unknown_col': {'S': 'not_strict'}}]
        db_result = {'Responses': {'autotest_dynamo_db': db_items}}

        self.dynamo_client.dynamo_client.batch_get_item = Mock(return_value=db_result)

        result = self.dynamo_client.batch_get_items_one_table(keys_list=[{'hash_col': 'b'}], fetch_all_fields=True)

        self.assertEqual(result, [{'hash_col': 'b', 'range_col': 10, 'unknown_col': 'not_strict'}])


    def test_get_by_query__max_items_and_count__raises(self):
        with self.assertRaises(Exception) as e:
            self.dynamo_client.get_by_query({'hash_col': 'key'}, table_name=self.table_name, max_items=3,
                                                           return_count=True)
        expected_msg = "DynamoDbCLient.get_by_query does not support `max_items` and `return_count` together"
        self.assertEqual(e.exception.args[0], expected_msg)


if __name__ == '__main__':
    unittest.main()



import unittest
import os

from unittest.mock import MagicMock
from unittest import mock


os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"


class siblings_TestCase(unittest.TestCase):

    CUSTOM_CONFIG = {
        "test": True
    }

    @mock.patch("boto3.client")
    def test_get_approximate_concurrent_executions(self, mock_boto_client):
        mock_get_metric_statistics_responses = [
            {
                'Datapoints':               [
                    {
                        'Average': 1.0,  # Extremely short - 1ms.
                        'Sum':     5  # Fake invocations
                    }
                ],
                'function_expected_result': 1
            },
            {
                'Datapoints':               [
                    {
                        'Average': 55000.0,  # A little shorter than 1 minute
                        'Sum':     5
                    }
                ],
                'function_expected_result': 1
            },
            {
                'Datapoints':               [
                    {
                        'Average': 65000.0,  # A little longer than 1 minute, should be ceiled to 2 minutes
                        'Sum':     5
                    }
                ],
                'function_expected_result': 2
            },
            {
                'Datapoints':               [
                    {
                        'Average': 120000.0,  # Exactly 2 minutes.
                        'Sum':     5
                    }
                ],
                'function_expected_result': 2
            },
            {
                'Datapoints':               [
                    {
                        'Average': 300000.0,  # Exactly 5 minutes, but a single invocation
                        'Sum':     1
                    }
                ],
                'function_expected_result': 1
            },
            {
                'Datapoints':               [
                    {
                        'Average': 300000.0,  # Exactly 5 minutes, invoked every minute
                        'Sum':     5
                    }
                ],
                'function_expected_result': 5
            },
        ]

        client = MagicMock()
        for experiment in mock_get_metric_statistics_responses:
            client.get_metric_statistics = MagicMock(return_value=experiment, side_effect=None)
            mock_boto_client.return_value = client

            # Reimport the component
            from sosw.components.siblings import SiblingsManager

            self.assertEqual(SiblingsManager(custom_config=self.CUSTOM_CONFIG).get_approximate_concurrent_executions(),
                             experiment['function_expected_result'])


    @mock.patch("boto3.client")
    def test_any_events_rules_enabled(self, mock_boto_client_v2):
        """
        Two functions:
        * my-test-func1 has only 1 rule DISABLED
        * my-test-func2 has 2 rules, one ENABLED, one DISABLED
        """


        def my_side_effect(**kwargs):

            if kwargs['Rule'] == 'test-rule-1':
                return {'Targets': [{'Arn': 'arn:aws:lambda:us-west-2:123:function:my-test-func1'}]}
            elif kwargs['Rule'] == 'test-rule-2':
                return {'Targets': [{'Arn': 'arn:aws:lambda:us-west-2:123:function:my-test-func2'}]}
            elif kwargs['Rule'] == 'test-rule-3':
                return {'Targets': [{'Arn': 'arn:aws:lambda:us-west-2:123:function:my-test-func2'}]}


        mock_list_rules = {
            'Rules': [{
                'Arn':                'arn:aws:events:us-west-2:123:rule/test-rule-1',
                'Description':        'Test rule',
                'Name':               'test-rule-1',
                'ScheduleExpression': 'rate(5 minutes)',
                'State':              'DISABLED'
            }, {
                'Arn':                'arn:aws:events:us-west-2:123:rule/test-rule-2',
                'Description':        'Test rule',
                'Name':               'test-rule-2',
                'ScheduleExpression': 'rate(5 minutes)',
                'State':              'DISABLED'
            },
                {
                    'Arn':                'arn:aws:events:us-west-2:123:rule/test-rule-3',
                    'Description':        'Test rule',
                    'Name':               'test-rule-3',
                    'ScheduleExpression': 'rate(5 minutes)',
                    'State':              'ENABLED'
                }]
        }

        client2 = MagicMock()

        client2.list_rules = MagicMock(return_value=mock_list_rules)
        client2.list_targets_by_rule = MagicMock(side_effect=my_side_effect)
        mock_boto_client_v2.return_value = client2

        # Reimport the component
        from sosw.components.siblings import SiblingsManager

        self.assertFalse(SiblingsManager(custom_config=self.CUSTOM_CONFIG).any_events_rules_enabled(type('lambda_context', (object,), {
            'invoked_function_arn': 'arn:aws:lambda:us-west-2:123:function:my-test-func1'
        })))

        self.assertTrue(SiblingsManager(custom_config=self.CUSTOM_CONFIG).any_events_rules_enabled(type('lambda_context', (object,), {
            'invoked_function_arn': 'arn:aws:lambda:us-west-2:123:function:my-test-func2'
        })))

        # testing auto_spawning defaults from config
        self.CUSTOM_CONFIG['auto_spawning'] = False
        self.assertFalse(SiblingsManager(custom_config=self.CUSTOM_CONFIG).any_events_rules_enabled(
            type('lambda_context', (object,), {
                'invoked_function_arn': 'arn:aws:lambda:us-west-2:123:function:my-test-func1'
            })))

        self.CUSTOM_CONFIG['auto_spawning'] = True
        self.assertTrue(SiblingsManager(custom_config=self.CUSTOM_CONFIG).any_events_rules_enabled(
            type('lambda_context', (object,), {
                'invoked_function_arn': 'arn:aws:lambda:us-west-2:123:function:my-test-func1'
            })))


    @mock.patch("boto3.client")
    def test_spawn_sibling(self, mock_boto_client):
        client = MagicMock()

        client.list_rules = MagicMock(return_value={})
        client.list_targets_by_rule = MagicMock(return_value={})

        mock_boto_client.return_value = client
        # Test here that sibling is spawned only if rule enabled.


if __name__ == '__main__':
    unittest.main()



import boto3
import os
import random
import unittest

from unittest.mock import MagicMock, patch

from sosw.orchestrator import Orchestrator
from sosw.labourer import Labourer
from sosw.test.variables import TEST_ORCHESTRATOR_CONFIG, TEST_TASK_CLIENT_CONFIG
from sosw.test.helpers_test import line_count


os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"


class Scheduler_IntegrationTestCase(unittest.TestCase):
    TEST_CONFIG = TEST_ORCHESTRATOR_CONFIG
    LABOURER = Labourer(id='some_function', arn='arn:aws:lambda:us-west-2:000000000000:function:some_function')


    def setUp(self):
        self.patcher = patch("sosw.app.get_config")
        self.get_config_patch = self.patcher.start()

        self.custom_config = self.TEST_CONFIG.copy()
        self.orchestrator = Orchestrator(self.custom_config)
        self.orchestrator.task_client.ecology_client = MagicMock()
        self.orchestrator.task_client.ecology_client.get_labourer_status.return_value = 4
        self.orchestrator.task_client.ecology_client.count_running_tasks_for_labourer.return_value = 0


    def tearDown(self):
        self.patcher.stop()

        try:
            del (os.environ['AWS_LAMBDA_FUNCTION_NAME'])
        except Exception:
            pass


    def test_call(self):
        self.orchestrator({'event': 42})

        some_labourer = self.orchestrator.task_client.register_labourers()[0]
        # self.orchestrator.task_client.ecology_client.count_running_tasks_for_labourer(some_labourer)



__all__ = ['benchmark']
__author__ = "Nikolay Grishchenko"
__version__ = "1.0"

import time


def benchmark(fn):
    """
    Decorator that should be used on class methods that you want to benchmark.
    It will aggregate to `self.stats` of the class timing of decorated functions.

    | `fn` - pointer to class function. Class is not yet initialized.
    | `self` - pointer to class instance. Passed during the call of decorated method.
    """


    def _timing_and_call_counter(self, *a, **kw):
        st = time.perf_counter()
        r = fn(self, *a, **kw)
        self.stats[f"time_{fn.__name__}"] += time.perf_counter() - st
        self.stats[f"calls_{fn.__name__}"] += 1
        return r


    return _timing_and_call_counter



"""
This class represents types of Workers.

We use a different name in order to avoid any possible mess with the class Worker (which the actual Workers
should inherit from). This one is just a set of settings and common methods for the type of Labourers.
"""

__all__ = ['Labourer']
__author__ = "Nikolay Grishchenko"

import logging
import time


logger = logging.getLogger()
logger.setLevel(logging.INFO)


class Labourer:
    ATTRIBUTES = ('id', 'arn')
    CUSTOM_ATTRIBUTES = ('arn', 'start', 'invoked', 'expired', 'health', 'health_metrics', 'average_duration',
                         'max_duration', 'max_attempts', 'max_simultaneous_invocations')
    id = None
    arn = None

    DEFAULTS = {
        'duration': 900,  # 15 minutes.
        'cooldown': 300,  # 5 minutes.
    }


    def __init__(self, **kwargs):

        if kwargs.pop('strict', False):
            for k, v in kwargs.items():
                if k in self.ATTRIBUTES:
                    setattr(self, k, v)
                else:
                    raise AttributeError(f"Not supported attribute for Labourer: {k}")
        else:
            for k, v in kwargs.items():
                setattr(self, k, v)

        self.set_defaults()


    def set_defaults(self):
        """
        Set some default values.
        You may (and should) overwrite them for your Labourers in config of the Orchestrator.
        """

        for k, v in self.DEFAULTS.items():
            if not getattr(self, k, None):
                setattr(self, k, v)


    def set_custom_attribute(self, name: str, value: int):
        """ Set timestamp attributes with some validation. Normally TaskManager is supposed to call me. """

        if name not in self.CUSTOM_ATTRIBUTES:
            raise ValueError(f"Failed to set custom attribute {name} with value {value} for Labourer {self.id}. "
                             f"Supported attributes are: {', '.join(self.CUSTOM_ATTRIBUTES)}.")
        logger.debug(f"Labourer {self.id} set custom attribute {name} with {value}")
        setattr(self, name, value)


    def get_attr(self, name: str):
        """ The Labourer must be first registered in TaskManager for this to work. """

        if name not in self.CUSTOM_ATTRIBUTES:
            raise ValueError(f"Supported values are: {', '.join(self.CUSTOM_ATTRIBUTES)}")

        try:
            return getattr(self, name)
        except AttributeError:
            raise AttributeError(f"The Labourer is not yet registered in TaskManager, and doesn't have any custom "
                                 f"attributes. Use TaskManager.register_labourer() first.")



from sosw.labourer import Labourer


TASKS_TABLE_CONFIG = {
    'row_mapper':       {
        'task_id':             'S',
        'labourer_id':         'S',
        'greenfield':          'N',
        'attempts':            'N',
        'closed_at':           'N',
        'completed_at':        'N',
        'desired_launch_time': 'N',
        'arn':                 'S',
        'payload':             'S'
    },
    'required_fields':  ['task_id', 'labourer_id', 'created_at', 'greenfield'],
    'table_name':       'autotest_sosw_tasks',
    'index_greenfield': 'autotest_sosw_tasks_greenfield',
    'field_names':      {
        'task_id':     'task_id',
        'labourer_id': 'labourer_id',
        'greenfield':  'greenfield',
    }
}

TEST_ECOLOGY_CLIENT_CONFIG = {
    'test': True
}

TEST_SIBLINGS_CLIENT_CONFIG = {
    'test': True
}

TEST_TASK_CLIENT_CONFIG = {
    'init_clients':                      [],
    'dynamo_db_config':                  TASKS_TABLE_CONFIG,
    'sosw_closed_tasks_table':           'autotest_sosw_closed_tasks',
    'sosw_retry_tasks_table':            'autotest_sosw_retry_tasks',
    'sosw_retry_tasks_greenfield_index': 'labourer_id_greenfield',
    'ecology_config':                    TEST_ECOLOGY_CLIENT_CONFIG,
    'labourers':                         {
        'some_function': {
            'arn':                          'arn:aws:lambda:us-west-2:000000000000:function:some_function',
            'max_simultaneous_invocations': 10,
        },
        1:               {'arn': 'bar'},
    },
}

TEST_ORCHESTRATOR_CONFIG = {
    'init_clients':   [],
    'task_config':    TEST_TASK_CLIENT_CONFIG,
    'ecology_config': TEST_ECOLOGY_CLIENT_CONFIG,
}

TEST_SCAVENGER_CONFIG = {
    'init_clients':     [],
    'dynamo_db_config': TASKS_TABLE_CONFIG,
    'ecology_config':   TEST_ECOLOGY_CLIENT_CONFIG,
    'task_config':      TEST_TASK_CLIENT_CONFIG,
}

TEST_SCHEDULER_CONFIG = {
    'init_clients':   [],
    'ecology_config': TEST_ECOLOGY_CLIENT_CONFIG,
    'siblings_config': TEST_SIBLINGS_CLIENT_CONFIG,
    'task_config':    TEST_TASK_CLIENT_CONFIG,
    'job_schema':     {
        'chunkable_attrs': [
            ('section', {}),
            ('store', {}),
            ('product', {}),
        ]
    }
}

TASKS = [
    {'task_id': '123', 'labourer_id': 'some_lambda', 'attempts': 3, 'greenfield': '123'},
    {'task_id': '124', 'labourer_id': 'another_lambda', 'attempts': 4, 'greenfield': '321'},
    {'task_id': '125', 'labourer_id': 'some_lambda', 'attempts': 3, 'greenfield': '123'}
]

LABOURERS = [Labourer(id='some_lambda', arn='some_arn', some_attr='yes'),
             Labourer(id='another_lambda', arn='another_arn'),
             Labourer(id='lambda3', arn='arn3')]

RETRY_TASKS = [
    {
        'task_id':             '1', 'labourer_id': 'some_function', 'greenfield': 1234, 'attempts': 2,
        'desired_launch_time': 7777, 'arn': 'some_arn', 'payload': {}
    },
    {
        'task_id':             '2', 'labourer_id': 'some_function', 'greenfield': 2345, 'attempts': 2,
        'desired_launch_time': 8888, 'arn': 'some_arn', 'payload': {}
    },
    # This one is not supposed to be launched
    {
        'task_id':             '3', 'labourer_id': 'some_function', 'greenfield': 3456, 'attempts': 3,
        'desired_launch_time': 9999, 'arn': 'some_arn', 'payload': {}
    },
]



import boto3
import datetime
import json
import logging
import os
import random
import re
import subprocess
import time
import unittest
import types

from copy import deepcopy
from pathlib import Path
import pprint
from unittest import mock
from unittest.mock import MagicMock, PropertyMock, patch

from sosw.scheduler import Scheduler, InvalidJob, global_vars
from sosw.labourer import Labourer
from sosw.components.helpers import chunks
from sosw.test.variables import TEST_SCHEDULER_CONFIG
from sosw.test.helpers_test import line_count

import sosw.scheduler as module


os.environ["STAGE"] = "test"
os.environ["autotest"] = "True"


class Scheduler_UnitTestCase(unittest.TestCase):
    TEST_CONFIG = TEST_SCHEDULER_CONFIG
    LABOURER = Labourer(id='some_function', arn='arn:aws:lambda:us-west-2:000000000000:function:some_function')
    FNAME = '/tmp/aglaya.txt'
    TODAY = datetime.date(2019, 1, 31)

    # Warning! Tthis Payload is not operational as it is. Should add `isolate_SOMETHING` in several places.
    PAYLOAD = {
        'sections': {
            'section_funerals':    {
                'stores': {
                    'store_flowers': None,
                    'store_caskets': None,
                },
            },
            'section_weddings':    {
                'stores': {
                    'store_flowers': None,
                    'store_limos':   None,
                    'store_music':   {
                        'products': ['product_march', 'product_chorus', 740, 'product,4', 'product 5'],
                    },
                }
            },
            'section_conversions': {
                'stores': {
                    'store_training':     {
                        'products': {
                            'product_history': None,
                            'product_prayer':  None,
                        }
                    },
                    'store_baptizing':    None,
                    'store_circumcision': None
                }
            },
            'section_gifts':       None
        }
    }


    def setUp(self):
        self.patcher = patch("sosw.app.get_config")
        self.get_config_patch = self.patcher.start()

        self.custom_config = deepcopy(self.TEST_CONFIG)
        self.custom_config['siblings_config'] = {
            'auto_spawning': True
        }

        lambda_context = types.SimpleNamespace()
        lambda_context.aws_request_id = 'AWS_REQ_ID'
        lambda_context.invoked_function_arn = 'arn:aws:lambda:us-west-2:000000000000:function:some_function'
        lambda_context.get_remaining_time_in_millis = MagicMock(side_effect=[100000, 100])
        global_vars.lambda_context = lambda_context
        self.custom_lambda_context = lambda_context  # This is to access from tests.

        with patch('boto3.client'):
            self.scheduler = module.Scheduler(self.custom_config)

        self.scheduler.s3_client = MagicMock()
        self.scheduler.sns_client = MagicMock()
        self.scheduler.task_client = MagicMock()
        self.scheduler.task_client.get_labourer.return_value = self.LABOURER
        self.scheduler.siblings_client = MagicMock()

        self.scheduler.st_time = time.time()


    def tearDown(self):
        self.patcher.stop()

        try:
            del (os.environ['AWS_LAMBDA_FUNCTION_NAME'])
        except Exception:
            pass

        for fname in [self.scheduler.local_queue_file, self.FNAME]:
            try:
                os.remove(fname)
            except Exception:
                pass


    def put_local_file(self, file_name=None, json=False):
        with open(file_name or self.scheduler.local_queue_file, 'w') as f:
            for x in range(10):
                if json:
                    f.write('{"key": "val", "number": "42", "boolean": true, "labourer_id": "some_function"}\n')
                else:
                    f.write(f"Hello Aglaya {x} {random.randint(0, 99)}\n")


    @staticmethod
    def line_count(file):
        return int(subprocess.check_output('wc -l {}'.format(file), shell=True).split()[0])


    def test_init__chunkable_attrs_not_end_with_s(self):
        config = self.custom_config
        config['job_schema']['chunkable_attrs'] = [('bad_name_ending_with_s', {})]

        with patch('boto3.client'):
            self.assertRaises(AssertionError, Scheduler, custom_config=config)


    def test_get_next_chunkable_attr(self):
        self.assertEqual(self.scheduler.get_next_chunkable_attr('store'), 'product')
        self.assertEqual(self.scheduler.get_next_chunkable_attr('stores'), 'product')
        self.assertEqual(self.scheduler.get_next_chunkable_attr('section'), 'store')
        self.assertIsNone(self.scheduler.get_next_chunkable_attr('product'))
        self.assertIsNone(self.scheduler.get_next_chunkable_attr('bad_name'))


    def test__queue_bucket(self):
        self.assertEqual(self.scheduler._queue_bucket, self.scheduler.config['queue_bucket'])


    def test__remote_queue_file(self):
        self.assertIn(f"{self.scheduler.config['s3_prefix'].strip('/')}", self.scheduler.remote_queue_file)
        self.assertIn(self.custom_lambda_context.aws_request_id, self.scheduler.remote_queue_file)


    def test__remote_queue_locked_file(self):
        self.assertIn(f"{self.scheduler.config['s3_prefix'].strip('/')}", self.scheduler.remote_queue_locked_file)
        self.assertIn('locked_', self.scheduler.remote_queue_locked_file)
        self.assertIn(self.custom_lambda_context.aws_request_id, self.scheduler.remote_queue_locked_file)


    ### Tests of file operations ###
    def test_pop_rows_from_file(self):
        self.put_local_file(self.FNAME)

        # Initial setup made 10 rows.
        self.assertEqual(self.line_count(self.FNAME), 10)

        # Pop a single top row
        self.scheduler.pop_rows_from_file(self.FNAME)
        self.assertEqual(self.line_count(self.FNAME), 9)

        # Pop multiple rows
        self.scheduler.pop_rows_from_file(self.FNAME, rows=5)
        self.assertEqual(self.line_count(self.FNAME), 4)

        # Catch StopIteration and return only remaining.
        r = self.scheduler.pop_rows_from_file(self.FNAME, rows=42)
        self.assertFalse(os.path.isfile(self.FNAME))
        self.assertEqual(len(r), 4)


    def test_pop_rows_from_file__reads_from_top(self):
        self.put_local_file(self.FNAME)

        r = self.scheduler.pop_rows_from_file(self.FNAME, rows=9)

        self.assertEqual(len(r), 9)
        self.assertTrue(r[0].startswith('Hello Aglaya 0'))

        with open(self.FNAME) as f:
            row = f.read()
            self.assertTrue(row.startswith('Hello Aglaya 9'))


    def test_pop_rows_from_file__missing_or_empty_file(self):
        # Missing file
        self.assertEqual(self.scheduler.pop_rows_from_file(self.FNAME), list())

        # Empty file
        Path(self.FNAME).touch()
        self.assertEqual(self.scheduler.pop_rows_from_file(self.FNAME), list())

        self.assertFalse(os.path.isfile(self.FNAME))


    def test_process_file(self):
        self.put_local_file(self.FNAME, json=True)
        self.scheduler.get_and_lock_queue_file = MagicMock(return_value=self.FNAME)
        self.scheduler.upload_and_unlock_queue_file = MagicMock()
        self.scheduler.task_client = MagicMock()
        self.scheduler.clean_tmp = MagicMock()

        with patch('sosw.scheduler.Scheduler._sleeptime_for_dynamo', new_callable=PropertyMock) as mock_sleeptime:
            mock_sleeptime.return_value = 0.0001

            self.scheduler.process_file()

            self.assertEqual(self.scheduler.task_client.create_task.call_count, 10)
            self.assertEqual(mock_sleeptime.call_count, 10)

            self.scheduler.upload_and_unlock_queue_file.assert_called_once()
            self.scheduler.clean_tmp.assert_called_once()
            # number of calls depends on the 'remaining_time_in_millis()' mock
            self.assertEqual(self.scheduler.siblings_client.spawn_sibling.call_count, 1)


    ### Tests of construct_job_data ###
    def test_construct_job_data(self):

        self.scheduler.chunk_dates = MagicMock(return_value=[{'a': 'foo'}, {'b': 'bar'}])
        self.scheduler.chunk_job = MagicMock()

        r = self.scheduler.construct_job_data({'pl': 1})

        self.scheduler.chunk_dates.assert_called_once()
        self.scheduler.chunk_job.assert_called()
        self.assertEqual(self.scheduler.chunk_job.call_count, 2)


    def test_construct_job_data__preserve_skeleton_through_chunkers(self):

        r = self.scheduler.construct_job_data({'pl': 1}, skeleton={'labourer_id': 'some'})
        print(r)

        for task in r:
            self.assertEqual(task['labourer_id'], 'some')


    def test_construct_job_data__empty_job(self):

        JOB = dict()
        r = self.scheduler.construct_job_data(JOB)
        self.assertEqual(r, [JOB])


    def test_construct_job_data__real_payload__for_debuging_logs(self):
        JOB = {
            'lambda_name':         'some_lambda',
            'period':              'last_2_days', 'isolate_days': True,
            'sections':            {
                '111': {'all_campaigns': True},
                '222': {'all_campaigns': True},
                '333': {
                    'isolate_stores': True,
                    'all_campaigns':  False,
                    'stores':         {'333-111': None, '333-222': None, '333-333': {'keep_me': 7}},
                }
            }, 'isolate_sections': 'True'
        }

        r = self.scheduler.construct_job_data(JOB)

        for t in r:
            print(t)

        self.assertEqual(len(r), 10)
        # self.assertEqual(1, 42)


    ### Tests of chunk_dates ###
    def test_chunk_dates(self):
        TESTS = [
            ({'period': 'today'}, 'today'),
            ({'period': 'yesterday'}, 'yesterday'),
            ({'period': 'last_3_days'}, 'last_x_days'),
            ({'period': '10_days_back'}, 'x_days_back'),
            ({'period': 'previous_2_days'}, 'previous_x_days'),
            ({'period': 'last_week'}, 'last_week')
        ]

        for test, func_name in TESTS:
            FUNCTIONS = ['today', 'yesterday', 'last_x_days', 'x_days_back', 'previous_x_days', 'last_week']
            for f in FUNCTIONS:
                setattr(self.scheduler, f, MagicMock())

            self.scheduler.chunk_dates(test)

            func = getattr(self.scheduler, func_name)
            func.assert_called_once()

            for bad_f_name in [x for x in FUNCTIONS if not x == func_name]:
                bad_f = getattr(self.scheduler, bad_f_name)
                bad_f.assert_not_called()


    def test_chunk_dates__preserve_skeleton(self):
        TESTS = [
            {'period': 'last_1_days', 'a': 'foo'},
            {'period': 'last_10_days', 'a': 'foo'},
            {'period': 'last_10_days', 'isolate_days': True, 'a': 'foo'},
            {'period': '1_days_back', 'a': 'foo'},
            {'period': '10_days_back', 'a': 'foo'},
            {'period': '10_days_back', 'isolate_days': True, 'a': 'foo'},  # Isolation here is abundant
        ]

        SKELETON = {'labourer_id': 'some'}

        for test in TESTS:
            if test.get('isolate_days'):
                pattern = '[a-z]+_([0-9]+)_days'
                try:
                    expected_number = int(re.match(pattern, test['period'])[1])
                except Exception:
                    expected_number = 1
            else:
                expected_number = 1

            r = self.scheduler.chunk_dates(job=test, skeleton=SKELETON)
            self.assertEqual(len(r), expected_number)
            for task in r:
                self.assertEqual(task['labourer_id'], 'some')


    def test_chunk_dates__preserve_skeleton__if_no_chunking(self):
        TASK = {
            'a': 'foo'
        }
        SKELETON = {'labourer_id': 'some'}

        r = self.scheduler.chunk_dates(job=TASK, skeleton=SKELETON)

        for task in r:
            self.assertEqual(task['labourer_id'], 'some')
            self.assertEqual(task['a'], 'foo')


    def test_chunk_dates__pops_period(self):
        TASK = {
            'period': 'last_42_days',
            'a':      'foo'
        }

        r = self.scheduler.chunk_dates(job=TASK)

        self.assertIn('period', TASK, "DANGER! Modified initial job!")
        for task in r:
            self.assertNotIn('period', task)
            self.assertEqual(task['a'], 'foo')


    def test_chunk_dates__last_x_days(self):

        TASK = {'period': 'last_5_days', 'isolate_days': True, 'a': 'foo'}
        today = datetime.date(2019, 1, 30)

        with patch('sosw.scheduler.datetime.date') as mdt:
            mdt.today.return_value = today
            r = self.scheduler.chunk_dates(TASK)

        self.assertEqual(len(r), 5)
        for i, task in enumerate(r):
            self.assertEqual(task['a'], 'foo')
            self.assertEqual(task['date_list'], [f"{self.TODAY - datetime.timedelta(days=6-i)}"])


    def test_chunk_dates__raises_invalid_period_pattern(self):
        TASK = {'period': 'putin_the_best'}
        self.assertRaises(ValueError, self.scheduler.chunk_dates, job=TASK), "Putin is not supported"


    def test_last_x_days(self):

        TESTS = [
            ('last_3_days', ['2019-01-27', '2019-01-28', '2019-01-29']),
            ('last_5_days', ['2019-01-25', '2019-01-26', '2019-01-27', '2019-01-28', '2019-01-29']),
        ]
        today = datetime.date(2019, 1, 30)

        with patch('sosw.scheduler.datetime.date') as mdt:
            mdt.today.return_value = today

            for test, expected in TESTS:
                self.assertEqual(self.scheduler.last_x_days(test), expected)


    def test_x_days_back(self):

        TESTS = [
            ('1_days_back', ['2019-01-29']),
            ('7_days_back', ['2019-01-23']),
            ('30_days_back', ['2018-12-31']),
        ]
        today = datetime.date(2019, 1, 30)

        with patch('sosw.scheduler.datetime.date') as mdt:
            mdt.today.return_value = today

            for test, expected in TESTS:
                self.assertEqual(self.scheduler.x_days_back(test), expected)

            last_week = self.scheduler.x_days_back('7_days_back')[0]
        self.assertEqual(today.weekday(), datetime.datetime.strptime(last_week, '%Y-%m-%d').weekday())


    def test_yesterday(self):

        TESTS = [
            ('yesterday', ['2019-04-10']),
        ]

        today = datetime.date(2019, 4, 11)

        with patch('sosw.scheduler.datetime.date') as mdt:
            mdt.today.return_value = today

            for test, expected in TESTS:
                self.assertEqual(self.scheduler.yesterday(test), expected)

    def test_today(self):
        TESTS = [
            ('today', ['2019-04-10']),
        ]
        today = datetime.date(2019, 4, 10)

        with patch('sosw.scheduler.datetime.date') as mdt:
            mdt.today.return_value = today

            for test, expected in TESTS:
                self.assertEqual(self.scheduler.today(test), expected)

    def test_previous_x_days(self):
        today = datetime.date(2019, 4, 30)

        TESTS = [
            ('previous_2_days', ['2019-04-26', '2019-04-27']),
            ('previous_3_days', ['2019-04-24', '2019-04-25', '2019-04-26'])
        ]

        with patch('sosw.scheduler.datetime.date') as mdt:
            mdt.today.return_value = today

            for test, expected in TESTS:
                self.assertEqual(self.scheduler.previous_x_days(test), expected)

    def test_last_week(self):
        today = datetime.date(2019, 4, 30)

        TESTS = [
            ('last_week', ['2019-04-21',
                           '2019-04-22',
                           '2019-04-23',
                           '2019-04-24',
                           '2019-04-25',
                           '2019-04-26',
                           '2019-04-27'])
        ]

        with patch('sosw.scheduler.datetime.date') as mdt:
            mdt.today.return_value = today

            for test, expected in TESTS:
                self.assertEqual(self.scheduler.last_week(test), expected)


    ### Tests of chunk_job ###
    def test_chunk_job__not_chunkable_config(self):
        self.scheduler.chunkable_attrs = []
        pl = deepcopy(self.PAYLOAD)

        r = self.scheduler.chunk_job(job=pl)
        # pprint.pprint(r)
        self.assertEqual(len(r), 1)
        self.assertEqual(r[0], pl)


    def test_chunk_job__not_raises_unchunkable_subtask__but_preserves_in_payload(self):
        pl = deepcopy(self.PAYLOAD)
        pl['sections']['section_conversions']['stores']['store_training']['isolate_products'] = True
        pl['sections']['section_conversions']['stores']['store_training']['products']['product_book'] = {
            'product_versions':
                {
                    'product_version_audio': None,
                    'product_version_paper': None,
                }
        }


        def find_product(t):
            try:
                return set(t['product_versions'].keys()) == {'product_version_audio', 'product_version_paper'}
            except Exception:
                return False


        # print(pl)
        r = self.scheduler.chunk_job(job=pl)
        # for t in r:
        #     print(t)

        self.assertTrue(any(find_product(task) for task in r))


    def test_chunk_job__raises__unsupported_vals__string(self):
        pl = deepcopy(self.PAYLOAD)

        pl['sections']['section_conversions']['isolate_stores'] = True
        pl['sections']['section_conversions']['stores']['store_training'] = 'some_string'

        self.assertRaises(InvalidJob, self.scheduler.chunk_job, job=pl)


    def test_chunk_job__raises__unsupported_vals__list_not_as_value(self):
        pl = deepcopy(self.PAYLOAD)
        pl['sections']['section_conversions']['isolate_stores'] = True
        pl['sections']['section_conversions']['stores']['store_training'] = ['just_a_string']

        self.assertRaises(InvalidJob, self.scheduler.chunk_job, job=pl)


    def test_chunk_job__not_raises__notchunkable__if_no_isolation(self):
        pl = deepcopy(self.PAYLOAD)

        pl['isolate_sections'] = True
        pl['sections']['section_conversions']['stores']['store_training'] = 'some_string'

        r = self.scheduler.chunk_job(job=pl)
        val = r[2]
        print(r)
        print(f"We chunked only first level (sections). The currently interesting is section #3, "
              f"where we put custom unchunkable payload: {val}")

        self.assertEqual(val['stores']['store_training'], 'some_string')


    def test_chunk_job(self):

        pl = deepcopy(self.PAYLOAD)
        pl['sections']['section_weddings']['stores']['store_music']['isolate_products'] = True
        pl['sections']['section_conversions']['stores']['store_training']['isolate_products'] = True

        response = self.scheduler.chunk_job(job=pl)

        # for row in response:
        #     pprint.pprint(row)
        #     print('\n')

        NUMBER_TASKS_EXPECTED = [
            ('sections', 'section_funerals', 1),
            ('sections', 'section_weddings', 7),
            ('sections', 'section_conversions', 4),
            ('stores', 'store_training', 2),
            ('stores', 'store_baptizing', 1),
            ('sections', 'section_gifts', 1),
        ]

        self.check_number_of_tasks(NUMBER_TASKS_EXPECTED, response)


    def test_chunk_job__unchunckable_preserve_custom_attrs(self):

        pl = {
            'sections': {
                'section_funerals': {'custom': 'data'},
                'section_weddings': None,
            }
        }

        response = self.scheduler.chunk_job(job=pl)
        # print(response)

        self.assertEqual([pl], response)


    def test_chunk_job__max_items_per_batch(self):
        """
        Tests that `max_products_per_batch` will actually make chunks of products of specific size.
        :return:
        """
        pl = deepcopy(self.PAYLOAD)
        pl['sections']['section_weddings']['stores']['store_music']['max_products_per_batch'] = 2

        response = self.scheduler.chunk_job(job=pl)

        NUMBER_TASKS_EXPECTED = [
            ('sections', 'section_weddings', 5),
        ]

        # for row in response:
        #     pprint.pprint(row)
        #     print('\n')

        self.check_number_of_tasks(NUMBER_TASKS_EXPECTED, response)
        batches = [x['products'] for x in response if x.get('stores') == ['store_music']]
        print(batches)

        self.assertEqual(batches,
                         list(chunks(pl['sections']['section_weddings']['stores']['store_music']['products'], 2)))


    ### Tests of other methods ###
    def test_extract_job_from_payload(self):

        TESTS = [
            ({'job': {'lambda_name': 'foo', 'payload_attr': 'val'}}, {'lambda_name': 'foo', 'payload_attr': 'val'}),
            ({'lambda_name': 'foo', 'payload_attr': 'val'}, {'lambda_name': 'foo', 'payload_attr': 'val'}),
            ({'lambda_name': 'arn:aws:lambda:us-west-2:000000000000:function:foo', 'payload_attr': 'val'},
             {'lambda_name': 'foo', 'payload_attr': 'val'}),
            ({'job': {'lambda_name': 'foo', 'payload_attr': 'val'}}, {'lambda_name': 'foo', 'payload_attr': 'val'}),

            # JSONs
            ('{"lambda_name": "foo", "payload_attr": "val"}', {'lambda_name': 'foo', 'payload_attr': 'val'}),
            ('{"job": {"lambda_name": "foo", "payload_attr": "val"}}', {'lambda_name': 'foo', 'payload_attr': 'val'}),
            ('{"job": "{\\"lambda_name\\": \\"foo\\", \\"payload_attr\\": \\"val\\"}"}',
             {'lambda_name': 'foo', 'payload_attr': 'val'}),

        ]

        for test, expected in TESTS:
            self.assertEqual(self.scheduler.extract_job_from_payload(test), expected)


    def test_extract_job_from_payload_raises(self):

        TESTS = [
            42,
            {'payload_attr': 'val'},
            "{'payload_attr': 'val'}",
            {'job': {'payload_attr': 'val'}},
            {"job": "bad one"},
        ]

        for test in TESTS:
            self.assertRaises(Exception, self.scheduler.extract_job_from_payload, test)


    def test_needs_chunking__isolate_root(self):

        pl = deepcopy(self.PAYLOAD)
        self.assertFalse(self.scheduler.needs_chunking('sections', pl))

        pl = deepcopy(self.PAYLOAD)
        pl['isolate_sections'] = True
        self.assertTrue(self.scheduler.needs_chunking('sections', pl))


    def test_needs_chunking__isolate_subdata(self):

        pl = deepcopy(self.PAYLOAD)
        pl['sections']['section_funerals']['isolate_stores'] = True

        self.assertTrue(self.scheduler.needs_chunking('sections', pl))
        self.assertTrue(self.scheduler.needs_chunking('stores', pl['sections']['section_funerals']))
        self.assertFalse(self.scheduler.needs_chunking('stores', pl['sections']['section_conversions']))


    def test_needs_chunking__isolate_subdata_deep(self):

        pl = deepcopy(self.PAYLOAD)
        pl['sections']['section_conversions']['stores']['store_training']['isolate_products'] = True
        # pprint.pprint(pl)

        self.assertFalse(self.scheduler.needs_chunking('stores', pl['sections']['section_funerals']))
        self.assertTrue(self.scheduler.needs_chunking('stores', pl['sections']['section_conversions']))
        self.assertTrue(self.scheduler.needs_chunking(
                'products', pl['sections']['section_conversions']['stores']['store_training']))
        self.assertTrue(self.scheduler.needs_chunking('sections', pl))


    def test_needs_chunking__max_items_per_batch(self):

        pl = deepcopy(self.PAYLOAD)

        # Verify that no chunking is required by default
        self.assertFalse(self.scheduler.needs_chunking('sections', pl))

        # Inject max_items_per_batch and recheck.
        pl['sections']['section_conversions']['stores']['store_training']['max_products_per_batch'] = 3
        self.assertTrue(self.scheduler.needs_chunking('sections', pl))


    def test_get_index_from_list(self):

        TESTS = [
            (0, 'a', ['a', 'b', 'c']),
            (0, 'name', ['names', 'b', 'c']),
            (2, 'c', ['a', 'b', 'c']),
            (1, 'b', {'a': 1, 'b': 2, 'c': 3}),
            (1, 'bob', {'a': 1, 'bobs': 2, 'c': 3}),
        ]

        for expected, attr, data in TESTS:
            self.assertEqual(expected, self.scheduler.get_index_from_list(attr, data))


    def check_number_of_tasks(self, expected_map, response):
        for key, val, expected in expected_map:
            r = filter(lambda task: task.get(key) == [val], response)
            # print(f"TEST OF FILTER: {t}: {len(list(t))}")
            self.assertEqual(len(list(r)), expected)


    def test_validate_list_of_vals(self):
        TESTS = [
            ({'a': None, 'b': None}, ['a', 'b']),
            (['a', 'b', 42], ['a', 'b', 42]),
            ([], []),
        ]

        for test, expected in TESTS:
            self.assertEqual(self.scheduler.validate_list_of_vals(test), expected)


    def test_get_and_lock_queue_file__s3_calls(self):

        self.scheduler.get_and_lock_queue_file()
        self.scheduler.s3_client.download_file.assert_called_once()
        self.scheduler.s3_client.copy_object.assert_called_once()
        self.scheduler.s3_client.delete_object.assert_called_once()
        self.scheduler.s3_client.upload_file.assert_not_called()


    def test_get_and_lock_queue_file__local_file_exists(self):

        with patch('os.path.isfile') as isfile_mock:
            isfile_mock.return_value = True

            r = self.scheduler.get_and_lock_queue_file()

        self.assertEqual(r, self.scheduler.local_queue_file)
        self.scheduler.s3_client.download_file.assert_not_called()
        self.scheduler.s3_client.copy_object.assert_not_called()
        self.scheduler.s3_client.delete_object.assert_not_called()

        self.scheduler.s3_client.upload_file.assert_called_once()


    def test_parse_job_to_file(self):

        SAMPLE_SIMPLE_JOB = {
            'lambda_name':  self.LABOURER.id,
            'some_payload': 'foo',
        }

        self.scheduler.parse_job_to_file(SAMPLE_SIMPLE_JOB)

        self.assertEqual(line_count(self.scheduler.local_queue_file), 1)

        with open(self.scheduler.local_queue_file, 'r') as f:
            row = json.loads(f.read())
            print(row)

            self.assertEqual(row['labourer_id'], self.LABOURER.id)
            self.assertEqual(row['some_payload'], 'foo')


    def test_parse_job_to_file__multiple_rows(self):

        SAMPLE_SIMPLE_JOB = {
            'lambda_name':      self.LABOURER.id,
            "isolate_sections": True,
            'sections':         {
                'section_technic':   None,
                'section_furniture': None,
            },
        }

        self.scheduler.parse_job_to_file(SAMPLE_SIMPLE_JOB)

        self.assertEqual(line_count(self.scheduler.local_queue_file), 2)

        with open(self.scheduler.local_queue_file, 'r') as f:
            for row in f.readlines():
                # print(row)
                parsed_row = json.loads(row)
                print(parsed_row)

                self.assertEqual(parsed_row['labourer_id'], self.LABOURER.id)
                self.assertEqual(len(parsed_row['sections']), 1)
                self.assertIn(parsed_row['sections'][0], SAMPLE_SIMPLE_JOB['sections'])


    def test_call__sample(self):
        SAMPLE_SIMPLE_JOB = {
            'lambda_name':  self.LABOURER.id,
            'some_payload': 'foo',
        }

        print(json.dumps(SAMPLE_SIMPLE_JOB))
        r = self.scheduler(json.dumps(SAMPLE_SIMPLE_JOB))
        print(r)

        self.scheduler.task_client.create_task.assert_called_once()

        self.scheduler.s3_client.download_file.assert_not_called()
        self.scheduler.s3_client.copy_object.assert_not_called()

        self.scheduler.s3_client.upload_file.assert_called_once()
        self.scheduler.s3_client.delete_object.assert_called_once()




import boto3
import datetime
import json
import logging
import os

from math import ceil
from sosw import Processor


__author__ = "Nikolay Grishchenko"
__email__ = "dev@bimpression.com"
__version__ = "1.00"
__license__ = "MIT"
__status__ = "Production"

__all__ = ['SiblingsManager']

logger = logging.getLogger()


class SiblingsManager(Processor):
    """
    This set of helpers can be used for Lambdas that want to invoke some siblings of self. Very useful for Lambdas
    processing queues and running out of time. Some good usecase you can find in the code of `es_ingest_us`

    The Role of your Lambda must have the following extra permissions to run correctly. Please note that we hardcode
    the Arn in the policy to avoid circular dependency when parsing YAML. This dependency is absolutely valid, but
    CloudFormation doesn't know how to parse it.

    | Read more:
    | https://aws.amazon.com/premiumsupport/knowledge-center/unable-validate-circular-dependency-cloudformation/

    .. code-block:: yaml

        Policies:
        - PolicyName: "YOUR_FUNCTION_NAME"
        PolicyDocument:
          Version: "2012-10-17"
          Statement:
          - Effect: "Allow"
            Action: "cloudwatch:GetMetricStatistics"
            Resource: "*"
          - Effect: "Allow"
            Action: "lambda:InvokeFunction"
            Resource: "arn:aws:lambda:us-west-2:737060422660:function:YOUR_FUNCTION_NAME"
    """

    DEFAULT_CONFIG = {
        'init_clients':  ['lambda', 'events', 'cloudwatch'],
        'auto_spawning': False
    }

    events_client: boto3.client = None
    lambda_client: boto3.client = None
    cloudwatch_client: boto3.client = None

    def any_events_rules_enabled(self, lambda_context):
        """
        Checks the `Status` of CloudWatch Events Rules.
        It is very important to use this checker before launching siblings.
        Otherwise, you can create an infinite autorespawning loop and waste **A LOT** of money.

        :param lambda_context:  Context object from your lambda_handler.

        :rtype: bool
        :raises ResourceNotFoundException: If Rule with the given `name` doesn't exist.
        """

        response = self.events_client.list_rules()
        logging.debug(lambda_context.invoked_function_arn)
        logging.debug(response)

        for rule in response.get('Rules', []):
            if not rule['State'] == 'ENABLED':
                continue

            targets = self.events_client.list_targets_by_rule(Rule=rule['Name']).get('Targets', [])
            logging.debug(targets)
            if any(t['Arn'] == lambda_context.invoked_function_arn for t in targets):
                logger.info(f"Function {lambda_context.invoked_function_arn} has at least one enabled rule: {rule}")
                return True

        return self.config['auto_spawning']


    def spawn_sibling(self, lambda_context, payload=None, force=False):
        """
        Asynchronously invokes a copy of same function to continue working.
        Should be called if there is still work left to do (ex: messages in the queue).

        Can optionally send some payload for example remaining unprocessed rows of something.
        Should be formatted as dictionary.

        .. :warning:
           Must recieve the global lambda_context set from lambda_handler.

        .. :warning:
           Very dangerous to use `force=True`! This can create infinite loops.
           Use only if you are sure what you are doing!

        :param lambda_context:  Context object from your lambda_handler.
        :param dict payload:    The payload to be put to event.
        :param bool force:      If specified True it will ignore the checks of enabled Events Rules.
        """

        payload = payload or {}
        if not isinstance(payload, str):
            payload = json.dumps(payload)

        name = os.environ.get('AWS_LAMBDA_FUNCTION_NAME', 'test_function')

        invocation_type = 'Event' if not os.environ.get('STAGE') == 'test' else 'DryRun'

        if not self.any_events_rules_enabled(lambda_context) and not force:
            logger.error("Can't call siblings because I don't find any enabled CloudWatch Rules for me.")
            return

        logger.info(f"Calling sibling of {name} with payload: {payload}")
        self.lambda_client.invoke(FunctionName=name, InvocationType=invocation_type, Payload=payload)


    def get_approximate_concurrent_executions(self, minutes_back=5, name=None):
        """
        Get approximate concurrent executions from CloudWatch Metrics.
        The value is **very** approximate and calculated as count of invocations during `minutes_back` divided
        by average duration in same period. Return value is rounded to integer using `ceil`.

        We assume that the Role has permissions to read CloudWatch.

        :param int minutes_back: Aggregate statistics for this number of minutes.
        :param str name: Name of the function to check. *Default: currently running lambda.*
        :rtype: int
        :return: Approximate number of concurrent executions.
        """

        name = name or os.environ.get('AWS_LAMBDA_FUNCTION_NAME', 'test_function')
        period = 60 * minutes_back
        st = datetime.datetime.now() - datetime.timedelta(minutes=minutes_back)
        en = datetime.datetime.now()

        # First we get the average duration.
        # Please notice that difference between StartTime and EndTime is exactly equal to `period`.
        # This makes sure that we get only one aggregated `Datapoint` in the result.
        response = self.cloudwatch_client.get_metric_statistics(
                Namespace='AWS/Lambda', MetricName="Duration",
                StartTime=st, EndTime=en, Period=period, Statistics=['Average'],
                Dimensions=[{"Name": "FunctionName", "Value": name}])

        # If we had invocations during `period` - we have just one `Datapoint` in response.
        if response['Datapoints']:
            average_duration = response['Datapoints'][0]['Average']

            # Now getting the number of invocations with same filters and period.
            response = self.cloudwatch_client.get_metric_statistics(
                    Namespace='AWS/Lambda', MetricName="Invocations",
                    StartTime=st, EndTime=en, Period=period, Statistics=['Sum'],
                    Dimensions=[{"Name": "FunctionName", "Value": name}])

            # We assume that if we have Duration, we definitely have invocations.
            # You may change this if you catch some VALID example where this is not True.
            number_of_invocations = response['Datapoints'][0]['Sum']

            return ceil(number_of_invocations * (average_duration / (1000 * period)))
        else:
            logger.info(f"Failed to get data from CW (or no data) for {name}.")
            return 0
